# -*- coding: utf-8 -*-

# system
import logging
import numpy as np
import os
import pickle
import time
import pkg_resources
import webbrowser
import yaml
from operator import itemgetter
from scipy.interpolate import interp1d

# Qt
from PyQt5.QtCore import (pyqtSignal, pyqtSlot, Qt, QTimer, QSize, QSettings,
                          QCoreApplication, QModelIndex)
from PyQt5.QtGui import QIcon, QKeySequence, QColor
from PyQt5.QtWidgets import (
    QApplication, QMainWindow,
    QWidget, QAction,
    QSlider, QLabel, QFrame,
    QPushButton, QComboBox,
    QListWidget, QListWidgetItem, QTreeView, QTreeWidget, QTreeWidgetItem,
    QAbstractItemView,
    QToolBar, QStatusBar, QProgressBar,
    QTextEdit, QFileDialog, QInputDialog,
    QVBoxLayout, QHBoxLayout, QMessageBox,
)

# pyqtgraph
import pyqtgraph as pg
from pyqtgraph.dockarea import DockArea

# vtk
vtk_error_msg = ""
try:
    import vtk

    from vtk import vtkRenderer
    from vtk import qt
    # import patched class that fixes scroll problem
    from .visualization import QVTKRenderWindowInteractor

    vtk_available = True
except ImportError as e:
    vtk_available = False
    vtk_error_msg = e
    vtkRenderer = None
    QVTKRenderWindowInteractor = None

# pymoskito
from .registry import get_registered_visualizers
from .simulation_interface import SimulatorInteractor, SimulatorView
from .visualization import MplVisualizer, VtkVisualizer, DummyVisualizer
from .processing_gui import PostProcessor
from .tools import (
    get_resource, PlainTextLogger, Exporter, create_button_from_action
)

__all__ = ["SimulationGui", "run"]


def run(regimes=None):
    """ Helper function to launch the PyMoskito GUI
    """
    app = QApplication([])
    prog = SimulationGui()
    if regimes is not None:
        prog.load_regimes_from_file(regimes)
    prog.show()
    app.exec_()


class SimulationGui(QMainWindow):
    """
    class for the graphical user interface
    """
    # TODO enable closing plot docks by right-clicking their name
    TABLEAU_COLORS = (
        ('blue', '#1f77b4'),
        ('orange', '#ff7f0e'),
        ('green', '#2ca02c'),
        ('red', '#d62728'),
        ('purple', '#9467bd'),
        ('brown', '#8c564b'),
        ('pink', '#e377c2'),
        ('gray', '#7f7f7f'),
        ('olive', '#bcbd22'),
        ('cyan', '#17becf'),
    )

    runSimulation = pyqtSignal()
    stopSimulation = pyqtSignal()
    playbackTimeChanged = pyqtSignal()
    regimeFinished = pyqtSignal()
    finishedRegimeBatch = pyqtSignal(bool)

    def __init__(self):
        # constructor of the base class
        QMainWindow.__init__(self)

        QCoreApplication.setOrganizationName("RST")
        QCoreApplication.setOrganizationDomain("https://tu-dresden.de/rst")
        QCoreApplication.setApplicationVersion(
            pkg_resources.require("PyMoskito")[0].version)
        QCoreApplication.setApplicationName(globals()["__package__"])

        # load settings
        self._settings = QSettings()
        self._init_settings()

        # initialize logger
        self._logger = logging.getLogger(self.__class__.__name__)

        # Create Simulation Backend
        self.guiProgress = None
        self.cmdProgress = None
        self.sim = SimulatorInteractor(self)
        self.runSimulation.connect(self.sim.run_simulation)
        self.stopSimulation.connect(self.sim.stop_simulation)
        self.sim.simulation_finalized.connect(self.new_simulation_data)
        self.currentDataset = None
        self.interpolator = None

        # sim setup viewer
        self.targetView = SimulatorView(self)
        self.targetView.setModel(self.sim.target_model)
        self.targetView.expanded.connect(self.target_view_changed)
        self.targetView.collapsed.connect(self.target_view_changed)

        # sim results viewer
        self.result_view = QTreeView()

        # the docking area allows to rearrange the user interface at runtime
        self.area = pg.dockarea.DockArea()

        # Window properties
        icon_size = QSize(25, 25)
        self.setCentralWidget(self.area)
        self.resize(1000, 700)
        self.setWindowTitle("PyMoskito")
        res_path = get_resource("mosquito.png")
        icon = QIcon(res_path)
        self.setWindowIcon(icon)

        # create docks
        self.propertyDock = pg.dockarea.Dock("Properties")
        self.animationDock = pg.dockarea.Dock("Animation")
        self.regimeDock = pg.dockarea.Dock("Regimes")
        self.lastSimDock = pg.dockarea.Dock("Previous Simulations")
        self.dataDock = pg.dockarea.Dock("Data & Plots")
        self.logDock = pg.dockarea.Dock("Log")

        # arrange docks
        self.area.addDock(self.animationDock, "right")
        self.area.addDock(self.lastSimDock, "left", self.animationDock)
        self.area.addDock(self.propertyDock, "bottom", self.lastSimDock)
        self.area.addDock(self.dataDock, "bottom", self.propertyDock)
        self.area.addDock(self.logDock, "bottom", self.dataDock)
        self.area.addDock(self.regimeDock, "left", self.lastSimDock)
        self.non_plotting_docks = list(self.area.findAll()[1].keys())

        self.standardDockState = self.area.saveState()

        # add widgets to the docks
        self.propertyDock.addWidget(self.targetView)

        # setup the visualizer
        self.visuComboBox = QComboBox()
        self.visualizer = None
        self._init_visualizer()

        # regime window
        self.regime_list = QListWidget(self)
        self.regime_list.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.regimeDock.addWidget(self.regime_list)
        self.regime_list.itemDoubleClicked.connect(self.regime_dclicked)
        self._regimes = []
        self.regime_file_name = ""

        self.actDeleteRegimes = QAction(self.regime_list)
        self.actDeleteRegimes.setText("&Delete Selected Regimes")
        # TODO shortcut works always, not only with focus on the regime list
        # self.actDeleteRegimes.setShortcutContext(Qt.WindowShortcut)
        self.actDeleteRegimes.setShortcut(QKeySequence(Qt.Key_Delete))
        self.actDeleteRegimes.triggered.connect(self.remove_regime_items)

        self.actSave = QAction(self)
        self.actSave.setText('Save Results As')
        self.actSave.setIcon(QIcon(get_resource("save.png")))
        self.actSave.setDisabled(True)
        self.actSave.setShortcut(QKeySequence.Save)
        self.actSave.triggered.connect(self.export_simulation_data)

        self.actLoadRegimes = QAction(self)
        self.actLoadRegimes.setText("Load Regimes from File")
        self.actLoadRegimes.setIcon(QIcon(get_resource("load.png")))
        self.actLoadRegimes.setDisabled(False)
        self.actLoadRegimes.setShortcut(QKeySequence.Open)
        self.actLoadRegimes.triggered.connect(self.load_regime_dialog)

        self.actExitOnBatchCompletion = QAction(self)
        self.actExitOnBatchCompletion.setText("&Exit On Batch Completion")
        self.actExitOnBatchCompletion.setCheckable(True)
        self.actExitOnBatchCompletion.setChecked(
            self._settings.value("control/exit_on_batch_completion") == "True"
        )
        self.actExitOnBatchCompletion.changed.connect(
            self.update_exit_on_batch_completion_setting)

        # regime management
        self.runningBatch = False
        self._current_regime_index = None
        self._current_regime_name = None

        self.regimeFinished.connect(self.run_next_regime)
        self.finishedRegimeBatch.connect(self.regime_batch_finished)

        # last sim window
        self.lastSimList = QListWidget(self)
        self.lastSimList.setFocusPolicy(Qt.NoFocus)
        self.lastSimDock.addWidget(self.lastSimList)
        self._lastSimulations = []
        self.lastSimList.itemDoubleClicked.connect(self.load_last_sim)

        # data window
        self.dataWidget = QWidget()
        self.dataLayout = QHBoxLayout()

        self.dataPointListWidget = QListWidget()
        self.dataPointListLayout = QVBoxLayout()
        self.dataPointListWidget.setLayout(self.dataPointListLayout)
        self.dataPointListWidget.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.dataLayout.addWidget(self.dataPointListWidget)

        self.dataPointManipulationLayout = QVBoxLayout()
        self.dataPointManipulationLayout.addStretch()
        self.dataPointManipulationLayout.setSpacing(5)
        self.dataPointManipulationWidget = QWidget()
        self.dataPointManipulationWidget.setLayout(self.dataPointManipulationLayout)
        self.dataLayout.addWidget(self.dataPointManipulationWidget)

        self.dataPointLabel = QLabel("Data", self)
        self.dataPointLabel.setAlignment(Qt.AlignCenter)
        self.dataPointManipulationLayout.addWidget(self.dataPointLabel)

        self.actAddDataPointToPlot = QAction(self)
        self.actAddDataPointToPlot.setText("&Add selected data to plot")
        self.actAddDataPointToPlot.setIcon(QIcon(get_resource("add.png")))
        self.actAddDataPointToPlot.setToolTip(
            "Add the selected data set from the left to the selected plot "
            "on the right.")
        self.actAddDataPointToPlot.triggered.connect(self.add_data_point_to_tree)
        self.dataPointManipulationLayout.addWidget(
            create_button_from_action(self.actAddDataPointToPlot))

        self.actExportDataPoint = QAction(self)
        self.actExportDataPoint.setText("&Export selected data")
        self.actExportDataPoint.setIcon(QIcon(get_resource("export.png")))
        self.actExportDataPoint.setToolTip(
            "Export the selected data set from the left to a csv or png file."
        )
        self.actExportDataPoint.triggered.connect(self.export_data_points_from_tree)
        self.dataPointManipulationLayout.addWidget(
            create_button_from_action(self.actExportDataPoint))

        self.actRemoveDataPointFromPlot = QAction(self)
        self.actRemoveDataPointFromPlot.setText("&Remove selected data from plot")
        self.actRemoveDataPointFromPlot.setIcon(QIcon(get_resource("delete.png")))
        self.actRemoveDataPointFromPlot.setToolTip(
            "Remove the selected data set from the plot on the right."
        )
        self.actRemoveDataPointFromPlot.triggered.connect(
            self.remove_data_point_from_tree)
        self.dataPointManipulationLayout.addWidget(
            create_button_from_action(self.actRemoveDataPointFromPlot))

        self.plotLabel = QLabel("Plots", self)
        self.plotLabel.setAlignment(Qt.AlignCenter)
        self.dataPointManipulationLayout.addWidget(self.plotLabel)

        self.actCreatePlot = QAction(self)
        self.actCreatePlot.setIcon(QIcon(get_resource("add.png")))
        self.actCreatePlot.setText("&Create a new plot")
        self.actCreatePlot.setToolTip("Create a new plot.")
        self.actCreatePlot.triggered.connect(self.add_plot_tree_item)
        self.dataPointManipulationLayout.addWidget(
            create_button_from_action(self.actCreatePlot))

        self.actShowPlot = QAction(self)
        self.actShowPlot.setText("&Show the selected plot")
        self.actShowPlot.setIcon(QIcon(get_resource("view.png")))
        self.actShowPlot.setToolTip(
            "Show the selected plot (double-click on name)."
        )
        self.actShowPlot.triggered.connect(self.show_selected_plot_tree_item)
        self.dataPointManipulationLayout.addWidget(
            create_button_from_action(self.actShowPlot))

        self.actDeletePlot = QAction(self)
        self.actDeletePlot.setText("&Delete selected plot")
        self.actDeletePlot.setIcon(QIcon(get_resource("delete.png")))
        self.actDeletePlot.setToolTip(
            "Delete the selected plot."
        )
        self.actDeletePlot.triggered.connect(self.remove_selected_plot_tree_items)
        self.dataPointManipulationLayout.addWidget(
            create_button_from_action(self.actDeletePlot))
        self.dataPointManipulationLayout.addStretch()

        self.dataPointTreeWidget = QTreeWidget()
        self.dataPointTreeWidget.setHeaderLabels(["PlotTitle", "DataPoint"])
        # self.dataPointTreeWidget.setSelectionMode(QAbstractItemView.MultiSelection)
        self.dataPointTreeWidget.itemDoubleClicked.connect(self.plot_vector_double_clicked)
        self.dataPointTreeWidget.setExpandsOnDoubleClick(0)
        self.dataPointTreeLayout = QVBoxLayout()

        self.dataPointTreeWidget.setLayout(self.dataPointTreeLayout)
        self.dataLayout.addWidget(self.dataPointTreeWidget)

        self.dataWidget.setLayout(self.dataLayout)
        self.dataDock.addWidget(self.dataWidget)

        # actions for simulation control
        self.actSimulateCurrent = QAction(self)
        self.actSimulateCurrent.setText("&Simulate Current Regime")
        self.actSimulateCurrent.setIcon(QIcon(get_resource("simulate.png")))
        self.actSimulateCurrent.setShortcut(QKeySequence("F5"))
        self.actSimulateCurrent.triggered.connect(self.start_simulation)

        self.actSimulateAll = QAction(self)
        self.actSimulateAll.setText("Simulate &All Regimes")
        self.actSimulateAll.setIcon(QIcon(get_resource("execute_regimes.png")))
        self.actSimulateAll.setShortcut(QKeySequence("F6"))
        self.actSimulateAll.setDisabled(True)
        self.actSimulateAll.triggered.connect(self.start_regime_execution)

        # actions for animation control
        self.actAutoPlay = QAction(self)
        self.actAutoPlay.setText("&Autoplay Simulation")
        self.actAutoPlay.setCheckable(True)
        self.actAutoPlay.setChecked(
            self._settings.value("control/autoplay_animation") == "True"
        )
        self.actAutoPlay.changed.connect(self.update_autoplay_setting)

        self.actPlayPause = QAction(self)
        self.actPlayPause.setText("Play Animation")
        self.actPlayPause.setIcon(QIcon(get_resource("play.png")))
        self.actPlayPause.setDisabled(True)
        self.actPlayPause.setShortcut(QKeySequence(Qt.Key_Space))
        self.actPlayPause.triggered.connect(self.start_animation)

        self.actStop = QAction(self)
        self.actStop.setText("Stop")
        self.actStop.setIcon(QIcon(get_resource("stop.png")))
        self.actStop.setDisabled(True)
        self.actStop.triggered.connect(self.stop_animation)

        self.actSlow = QAction(self)
        self.actSlow.setText("Slowest")
        self.actSlow.setIcon(QIcon(get_resource("slow.png")))
        self.actSlow.setDisabled(False)
        self.actSlow.triggered.connect(self.set_slowest_playback_speed)

        self.actFast = QAction(self)
        self.actFast.setText("Fastest")
        self.actFast.setIcon(QIcon(get_resource("fast.png")))
        self.actFast.setDisabled(False)
        self.actFast.triggered.connect(self.set_fastest_playback_speed)

        self.speedControl = QSlider(Qt.Horizontal, self)
        self.speedControl.setMaximumSize(200, 25)
        self.speedControl.setTickPosition(QSlider.TicksBothSides)
        self.speedControl.setDisabled(False)
        self.speedControl.setMinimum(0)
        self.speedControl.setMaximum(12)
        self.speedControl.setValue(6)
        self.speedControl.setTickInterval(6)
        self.speedControl.setSingleStep(2)
        self.speedControl.setPageStep(3)
        self.speedControl.valueChanged.connect(self.update_playback_speed)

        self.timeSlider = QSlider(Qt.Horizontal, self)
        self.timeSlider.setMinimum(0)
        self.timeSliderRange = 1000
        self.timeSlider.setMaximum(self.timeSliderRange)
        self.timeSlider.setTickInterval(1)
        self.timeSlider.setTracking(True)
        self.timeSlider.setDisabled(True)
        self.timeSlider.valueChanged.connect(self.update_playback_time)

        self.playbackTime = 0
        self.playbackGain = 1
        self.currentStepSize = 0
        self.currentStartTime = 0
        self.currentEndTime = 0
        self.playbackTimer = QTimer()
        self.playbackTimer.timeout.connect(self.increment_playback_time)
        self.playbackTimeChanged.connect(self.update_gui)
        self.playbackTimeout = 33  # in [ms] -> 30 fps

        self.actResetCamera = QAction(self)
        self.actResetCamera.setText("Reset Camera")
        self.actResetCamera.setIcon(QIcon(get_resource("video_camera.png")))
        self.actResetCamera.setDisabled(True)
        self.actResetCamera.setEnabled(self.visualizer.can_reset_view)
        self.actResetCamera.triggered.connect(self.reset_camera_clicked)

        # postprocessing
        self.actPostprocessing = QAction(self)
        self.actPostprocessing.setText("Launch Postprocessor")
        self.actPostprocessing.setIcon(QIcon(get_resource("processing.png")))
        self.actPostprocessing.setDisabled(False)
        self.actPostprocessing.triggered.connect(self.postprocessing_clicked)
        self.actPostprocessing.setShortcut(QKeySequence("F7"))

        self.postprocessor = None

        # toolbar
        self.toolbarSim = QToolBar("Simulation")
        self.toolbarSim.setContextMenuPolicy(Qt.PreventContextMenu)
        self.toolbarSim.setMovable(False)
        self.toolbarSim.setIconSize(icon_size)
        self.addToolBar(self.toolbarSim)
        self.toolbarSim.addAction(self.actLoadRegimes)
        self.toolbarSim.addAction(self.actSave)
        self.toolbarSim.addSeparator()
        self.toolbarSim.addAction(self.actSimulateCurrent)
        self.toolbarSim.addAction(self.actSimulateAll)
        self.toolbarSim.addSeparator()
        self.toolbarSim.addAction(self.actPlayPause)
        self.toolbarSim.addAction(self.actStop)
        self.toolbarSim.addWidget(self.timeSlider)
        self.toolbarSim.addSeparator()
        self.toolbarSim.addAction(self.actSlow)
        self.toolbarSim.addWidget(self.speedControl)
        self.toolbarSim.addAction(self.actFast)
        self.toolbarSim.addSeparator()
        self.toolbarSim.addAction(self.actResetCamera)
        self.toolbarSim.addWidget(self.visuComboBox)
        self.toolbarSim.addSeparator()
        self.toolbarSim.addAction(self.actPostprocessing)
        self.postprocessor = None

        # log dock
        self.logBox = QTextEdit(self)
        self.logBox.setReadOnly(True)
        self.logBox.setLineWrapMode(QTextEdit.NoWrap)
        self.logBox.ensureCursorVisible()
        self.logDock.addWidget(self.logBox)

        # init logger for logging box
        self.textLogger = PlainTextLogger(self._settings,
                                          logging.INFO)
        self.textLogger.set_target_cb(self.logBox)
        logging.getLogger().addHandler(self.textLogger)

        # menu bar
        fileMenu = self.menuBar().addMenu("&File")
        fileMenu.addAction(self.actLoadRegimes)
        fileMenu.addAction(self.actSave)
        fileMenu.addAction("&Quit", self.close, QKeySequence(Qt.CTRL + Qt.Key_W))

        editMenu = self.menuBar().addMenu("&Edit")
        editMenu.addAction(self.actDeleteRegimes)

        self.viewMenu = self.menuBar().addMenu('&View')
        self.actLoadStandardState = QAction('&Restore Default View')
        self.viewMenu.addAction(self.actLoadStandardState)
        self.actLoadStandardState.triggered.connect(self.loadStandardDockState)
        self.actShowCoords = QAction("&Show Coordinates", self)
        self.actShowCoords.setCheckable(True)
        self.actShowCoords.setChecked(
            self._settings.value("view/show_coordinates") == "True"
        )
        self.viewMenu.addAction(self.actShowCoords)
        self.actShowCoords.changed.connect(self.update_show_coords_setting)

        simMenu = self.menuBar().addMenu("&Simulation")
        simMenu.addAction(self.actSimulateCurrent)
        simMenu.addAction(self.actSimulateAll)
        simMenu.addAction(self.actExitOnBatchCompletion)
        simMenu.addAction(self.actPostprocessing)

        animMenu = self.menuBar().addMenu("&Animation")
        animMenu.addAction(self.actPlayPause)
        animMenu.addAction("&Increase Playback Speed",
                           self.increment_playback_speed,
                           QKeySequence(Qt.CTRL + Qt.Key_Plus))
        animMenu.addAction("&Decrease Playback Speed",
                           self.decrement_playback_speed,
                           QKeySequence(Qt.CTRL + Qt.Key_Minus))
        animMenu.addAction("&Reset Playback Speed",
                           self.reset_playback_speed,
                           QKeySequence(Qt.CTRL + Qt.Key_0))
        animMenu.addAction(self.actAutoPlay)
        animMenu.addAction(self.actResetCamera)

        plotMenu = self.menuBar().addMenu("&Plotting")
        plotMenu.addAction(self.actAddDataPointToPlot)
        plotMenu.addAction(self.actExportDataPoint)
        plotMenu.addAction(self.actRemoveDataPointFromPlot)
        plotMenu.addAction(self.actCreatePlot)
        plotMenu.addAction(self.actShowPlot)
        plotMenu.addAction(self.actDeletePlot)

        helpMenu = self.menuBar().addMenu("&Help")
        helpMenu.addAction("&Online Documentation", self.show_online_docs)
        helpMenu.addAction("&About", self.show_info)

        # status bar
        self.status = QStatusBar(self)
        self.setStatusBar(self.status)
        self.statusLabel = QLabel("Idle")
        self.statusBar().addPermanentWidget(self.statusLabel)
        self.timeLabel = QLabel("t=0.0")
        self.statusBar().addPermanentWidget(self.timeLabel)
        self.coordLabel = QLabel("x=0.0 y=0.0")
        self.statusBar().addPermanentWidget(self.coordLabel)

        self._logger.info("Simulation GUI is up and running.")

    def _init_visualizer(self):
        """ Initialize the visualizer combobox """
        self.visuComboBox.currentTextChanged.connect(self._visualizer_changed)
        self.visuComboBox.addItem("None")

        if not vtk_available:
            self._logger.warning(
                "loading vtk failed with:{}".format(vtk_error_msg))

        # get all registered visualizers
        available_vis = get_registered_visualizers()

        # fill the combo box
        for cls, name in available_vis:
            self.visuComboBox.addItem(name)

        # instantiate the first visualizer or None
        if self.visuComboBox.count() > 0:
            self.visuComboBox.setCurrentIndex(1)
        else:
            self.visuComboBox.setCurrentIndex(0)

    @pyqtSlot(str)
    def _visualizer_changed(self, name):
        """ Slot to update the visualizer """
        self._logger.info(
            "selecting visualizer '{}'".format(name))
        available_vis = get_registered_visualizers()
        cls = next((_c for _c, _n in available_vis if _n == name),
                   DummyVisualizer)
        self._update_visualizer(cls)

    def _update_visualizer(self, visu_cls):
        """ Update the current visualizer """
        if self.animationDock.widgets:
            self.animationDock.layout.removeWidget(
                self.animationDock.widgets[0])
            self.animationDock.widgets[0].deleteLater()
            del self.animationDock.widgets[0]

        self.animationLayout = QVBoxLayout()
        if visu_cls is DummyVisualizer:
            self.visualizer = DummyVisualizer()
            # self.animationWidget = QWidget()
            # self.animationDock.addWidget(self.animationWidget, row=0)
        elif issubclass(visu_cls, MplVisualizer):
            self.animationWidget = QWidget()
            self.visualizer = visu_cls(self.animationWidget,
                                       self.animationLayout)
            self.animationDock.addWidget(self.animationWidget, row=0)
        elif issubclass(visu_cls, VtkVisualizer):
            if vtk_available:
                # vtk window
                self.animationFrame = QFrame()
                self.vtkWidget = QVTKRenderWindowInteractor(
                    self.animationFrame)
                self.animationLayout.addWidget(self.vtkWidget)
                self.animationFrame.setLayout(self.animationLayout)
                self.animationDock.addWidget(self.animationFrame, row=0)
                self.vtk_renderer = vtkRenderer()
                self.vtkWidget.GetRenderWindow().AddRenderer(
                    self.vtk_renderer)
                self.visualizer = visu_cls(self.vtk_renderer)
                self.vtkWidget.Initialize()
            else:
                self._logger.warning("visualizer depends on vtk which is "
                                     "not available on this system!")
        else:
            raise NotImplementedError("Unsupported visualizer type '{}'"
                                      "".format(visu_cls))
        if hasattr(self, "actResetCamera"):
            self.actResetCamera.setEnabled(self.visualizer.can_reset_view)

    def add_plot_tree_item(self, default=False):
        text = "plot_{:03d}".format(self.dataPointTreeWidget.topLevelItemCount())
        if not default:
            name, ok = QInputDialog.getText(self,
                                            "Select a plot title",
                                            "name of the new plot:",
                                            text=text)
            if not (ok and name):
                return
        else:
            name = text

        similar_items = self.dataPointTreeWidget.findItems(name,
                                                           Qt.MatchExactly)
        if similar_items:
            self._logger.error("A plot with name '{}' already exists"
                               "".format(name))
            return

        if name in self.non_plotting_docks:
            self._logger.error("Name '{}' not allowed for a plot window "
                               "since it is already taken by another dock."
                               "".format(name))
            return

        top_level_item = QTreeWidgetItem()
        top_level_item.setText(0, name)
        self.dataPointTreeWidget.addTopLevelItem(top_level_item)
        top_level_item.setExpanded(1)
        self.dataPointTreeWidget.setCurrentItem(top_level_item)
        self.plot_vector_double_clicked(top_level_item)

    def show_selected_plot_tree_item(self):
        items = self.dataPointTreeWidget.selectedItems()
        for item in items:
            self.show_plot_tree_item(item)

    def remove_selected_plot_tree_items(self):
        items = self.dataPointTreeWidget.selectedItems()
        if not items:
            self._logger.error("Can't remove plot: no plot selected.")
            return

        for item in items:
            self.remove_plot_tree_item(item)

    def remove_plot_tree_item(self, item):
        # get the  top item
        while item.parent():
            item = item.parent()

        text = "The marked plot '" + item.text(0) + "' will be deleted!"
        reply_button = QMessageBox.warning(self, "Plot delete", text,
                                          QMessageBox.Ok | QMessageBox.Cancel)
        if reply_button == QMessageBox.Ok:
            open_docks = [dock.title() for dock in self.find_all_plot_docks()]
            if item.text(0) in open_docks:
                self.area.docks[item.text(0)].close()

            self.dataPointTreeWidget.takeTopLevelItem(
                self.dataPointTreeWidget.indexOfTopLevelItem(item))

    def add_data_point_to_tree(self):
        if not self.dataPointListWidget.selectedIndexes():
            self._logger.error("Can't add data set: no data set selected.")
            return

        data_points = []
        for item in self.dataPointListWidget.selectedItems():
            data_points.append(item.text())

        top_level_items = self.dataPointTreeWidget.selectedItems()
        if not top_level_items:
            if self.dataPointTreeWidget.topLevelItemCount() < 2:
                if self.dataPointTreeWidget.topLevelItemCount() < 1:
                    self.add_plot_tree_item(default=True)
                top_level_item = self.dataPointTreeWidget.topLevelItem(0)
            else:
                self._logger.error("Can't add data set: no plot selected.")
                return
        else:
            top_level_item = top_level_items[0]

        while top_level_item.parent():
            top_level_item = top_level_item.parent()

        top_level_item_list = []
        for i in range(top_level_item.childCount()):
            top_level_item_list.append(top_level_item.child(i).text(1))

        dock = next((d for d in self.find_all_plot_docks()
                     if d.title() == top_level_item.text(0)), None)

        for data_point in data_points:
            if data_point not in top_level_item_list:
                child = QTreeWidgetItem()
                child.setText(1, data_point)

                color = self._get_color(top_level_item.childCount())
                child.setBackground(0, color)

                top_level_item.addChild(child)

                if dock:
                    widget = dock.widgets[0]
                    self.plot_data_vector_member(child, widget)
            else:
                self._logger.error("Can't add data set: "
                                   "Set '{}' is already present selected plot"
                                   "".format(data_point))

    def export_data_points_from_tree(self):
        data_points = {}
        for item in self.dataPointListWidget.selectedItems():
            data_points[item.text()] = self._get_data_by_name(item.text())

        if not data_points:
            self._logger.error("Data export failed: No data selected.")
            return

        self.export(data_points)

    def remove_data_point_from_tree(self):
        items = self.dataPointTreeWidget.selectedItems()
        if not items:
            self._logger.error("Can't remove data set: no set selected.")
            return

        top_item = items[0]
        while top_item.parent():
            top_item = top_item.parent()

        top_item.takeChild(top_item.indexOfChild(items[0]))

        for i in range(top_item.childCount()):
            color_item = self._get_color(i)
            top_item.child(i).setBackground(0, color_item)

        self._update_plot(top_item)

    def _get_color(self, idx):
        sat_idx = idx % len(self.TABLEAU_COLORS)
        color = QColor(self.TABLEAU_COLORS[sat_idx][1])
        return color

    def _get_pen(self, idx):
        pen = pg.mkPen(self._get_color(idx), width=2)
        return pen

    def plot_vector_double_clicked(self, item):
        # check if a top level item has been clicked
        if not item.parent():
            self.show_plot_tree_item(item)

    def show_plot_tree_item(self, item):
        title = item.text(0)
        # check if plot has already been opened
        open_docks = [dock.title() for dock in self.find_all_plot_docks()]
        if title in open_docks:
            # update and raise the dock
            self._update_plot(item)
            try:
                self.area.docks[title].raiseDock()
            except:
                pass
        else:
            # create a new dock
            self.plot_data_vector(item)

    def load_last_sim(self, item):
        sim_name = str(item.text())
        try:
            idx = self.lastSimList.row(item)
        except ValueError as e:
            self._logger.error("load_last_sim(): No results called "
                               "'{0}'".format(sim_name))
            return False

        if idx >= len(self._lastSimulations):
            self._logger.error("load_last_sim(): Invalid index '{}')".format(idx))
            return False

        self._logger.info("restoring simulation '{}'".format(sim_name))

        self.currentDataset = self._lastSimulations[idx]
        if self._lastSimulations[idx]:
            self._read_results()
            self._update_data_list()
            self._update_plots()
            lsettings = self.currentDataset["modules"]
            self.visualizer.update_config(lsettings)
            lsettings["clear previous"] = True
            self.sim.restore_regime(lsettings)
            self.update_gui()

        self.setQListItemBold(self.lastSimList, item)
        self.setQListItemBold(self.regime_list, item)
        self.statusBar().showMessage(
            "restored simulation '{}'.".format(sim_name),
            1000)

    def _add_setting(self, setting, value):
        """
        Add a setting, if settings is present, no changes are made.

        Args:
            setting(str): Setting to add.
            value: Value to be set.
        """
        if not self._settings.contains(setting):
            self._settings.setValue(setting, value)

    def _init_settings(self):
        """
        Provide initial settings for the config file.

        """
        # path management
        self._add_setting("path/simulation_results",
                          os.path.join(os.path.curdir,
                                       "results",
                                       "simulation"))
        self._add_setting("path/simulation_results",
                          os.path.join(os.path.curdir,
                                       "results",
                                       "simulation"))
        self._add_setting("path/postprocessing_results",
                          os.path.join(os.path.curdir,
                                       "results",
                                       "postprocessing"))
        self._add_setting("path/metaprocessing_results",
                          os.path.join(os.path.curdir,
                                       "results",
                                       "metaprocessing"))
        self._add_setting("path/previous_plot_export", os.path.curdir)
        self._add_setting("path/previous_plot_format", ".csv")

        # control flow management
        self._add_setting("control/autoplay_animation", "False")
        self._add_setting("control/exit_on_batch_completion", "False")

        # view management
        self._add_setting("view/show_coordinates", "True")

        # log management
        self._add_setting("log_colors/CRITICAL", "#DC143C")
        self._add_setting("log_colors/ERROR", "#B22222")
        self._add_setting("log_colors/WARNING", "#DAA520")
        self._add_setting("log_colors/INFO", "#101010")
        self._add_setting("log_colors/DEBUG", "#4682B4")
        self._add_setting("log_colors/NOTSET", "#000000")

        self._add_setting("view/show_time_on_export", "False")
        self._add_setting("view/export_width", "800")
        self._add_setting("view/export_height", "600")

    def _write_settings(self):
        """ Store the application state. """
        pass

    @pyqtSlot()
    def update_autoplay_setting(self):
        self._settings.setValue("control/autoplay_animation",
                                str(self.actAutoPlay.isChecked()))

    @pyqtSlot()
    def update_show_coords_setting(self):
        self._settings.setValue("view/show_coordinates",
                                str(self.actShowCoords.isChecked()))

    @pyqtSlot()
    def update_exit_on_batch_completion_setting(self, state=None):
        if state is None:
            state = self.actExitOnBatchCompletion.isChecked()
        self._settings.setValue("control/exit_on_batch_completion", str(state))

    def set_visualizer(self, vis):
        self.visualizer = vis
        self.vtkWidget.Initialize()

    @pyqtSlot()
    def start_animation(self):
        """
        start the animation
        """
        self._logger.info("Starting Playback")

        # if we are at the end, start from the beginning
        if self.playbackTime >= self.currentEndTime:
            self.playbackTime = self.currentStartTime
            self.playbackTimeChanged.emit()

        self.actPlayPause.setText("Pause Animation")
        self.actPlayPause.setIcon(QIcon(get_resource("pause.png")))
        self.actPlayPause.triggered.disconnect(self.start_animation)
        self.actPlayPause.triggered.connect(self.pause_animation)
        self.playbackTimer.start(self.playbackTimeout)

    @pyqtSlot()
    def pause_animation(self):
        """
        pause the animation
        """
        self._logger.info("Pausing Playback")
        self.playbackTimer.stop()
        self.actPlayPause.setText("Play Animation")
        self.actPlayPause.setIcon(QIcon(get_resource("play.png")))
        self.actPlayPause.triggered.disconnect(self.pause_animation)
        self.actPlayPause.triggered.connect(self.start_animation)

    def stop_animation(self):
        """
        Stop the animation if it is running and reset the playback time.
        """
        self._logger.info("Stopping Playback")
        if self.actPlayPause.text() == "Pause Animation":
            # animation is playing -> stop it
            self.playbackTimer.stop()
            self.actPlayPause.setText("Play Animation")
            self.actPlayPause.setIcon(QIcon(get_resource("play.png")))
            self.actPlayPause.triggered.disconnect(self.pause_animation)
            self.actPlayPause.triggered.connect(self.start_animation)

        self.playbackTime = self.currentStartTime
        self.playbackTimeChanged.emit()

    @pyqtSlot()
    def start_simulation(self):
        """
        start the simulation and disable start button
        """
        if self._current_regime_index is None:
            regime_name = ""
        else:
            regime_name = str(self.regime_list.item(
                self._current_regime_index).text())

        self.statusLabel.setText("simulating {}".format(regime_name))
        self._logger.info("Simulating: {}".format(regime_name))

        self.actSimulateCurrent.setIcon(QIcon(
            get_resource("stop_simulation.png")))
        self.actSimulateCurrent.setText("Abort &Simulation")
        self.actSimulateCurrent.triggered.disconnect(self.start_simulation)
        self.actSimulateCurrent.triggered.connect(self.stop_simulation)

        if not self.runningBatch:
            self.actSimulateAll.setDisabled(True)

        self.guiProgress = QProgressBar(self)
        self.sim.simulationProgressChanged.connect(self.guiProgress.setValue)
        self.statusBar().addWidget(self.guiProgress)
        self.runSimulation.emit()

    @pyqtSlot()
    def stop_simulation(self):
        self.stopSimulation.emit()

    def export_simulation_data(self, ok):
        """
        Query the user for a custom name and export the current simulation
        results.

        :param ok: unused parameter from QAction.triggered() Signal
        """
        self._save_data()

    def _save_data(self, file_path=None):
        """
        Save the current simulation results.

        If *fie_name* is given, the result will be saved to the specified
        location, making automated exporting easier.

        Args:
            file_path(str): Absolute path of the target file. If `None` the
                use will be asked for a storage location.
        """
        regime_name = self._regimes[self._current_regime_index]["Name"]

        if file_path is None:
            # get default path
            path = self._settings.value("path/simulation_results")

            # create canonic file name
            suggestion = self._simfile_name(regime_name)
        else:
            path = os.path.dirname(file_path)
            suggestion = os.path.basename(file_path)

        # check if path exists otherwise create it
        if not os.path.isdir(path):
            box = QMessageBox()
            box.setText("Export Folder does not exist yet.")
            box.setInformativeText("Do you want to create it? \n"
                                   "{}".format(os.path.abspath(path)))
            box.setStandardButtons(QMessageBox.Ok | QMessageBox.No)
            box.setDefaultButton(QMessageBox.Ok)
            ret = box.exec_()
            if ret == QMessageBox.Ok:
                os.makedirs(path)
            else:
                path = os.path.abspath(os.path.curdir)
                file_path = None

        # If no path was given, present the default and let the user choose
        if file_path is None:
            dialog = QFileDialog(self)
            dialog.setAcceptMode(QFileDialog.AcceptSave)
            dialog.setFileMode(QFileDialog.AnyFile)
            dialog.setDirectory(path)
            dialog.setNameFilter("PyMoskito Results (*.pmr)")
            dialog.selectFile(suggestion)

            if dialog.exec_():
                file_path = dialog.selectedFiles()[0]
            else:
                self._logger.warning("Export Aborted")
                return -1

        # ask whether this should act as new default
        path = os.path.abspath(os.path.dirname(file_path))
        if path != self._settings.value("path/simulation_results"):
            box = QMessageBox()
            box.setText("Use this path as new default?")
            box.setInformativeText("{}".format(path))
            box.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            box.setDefaultButton(QMessageBox.Yes)
            ret = box.exec_()
            if ret == QMessageBox.Yes:
                self._settings.setValue("path/simulation_results", path)

        self.currentDataset.update({"regime name": regime_name})
        with open(file_path, "wb") as f:
            pickle.dump(self.currentDataset, f, protocol=4)

        self.statusLabel.setText("results saved to {}".format(file_path))
        self._logger.info("results saved to {}".format(file_path))

    def _simfile_name(self, regime_name):
        """ Create a canonical name for a simulation result file
        """
        suggestion = (time.strftime("%Y%m%d-%H%M%S")
                      + "_" + regime_name + ".pmr")
        return suggestion

    def load_regime_dialog(self):
        regime_path = os.path.join(os.curdir)

        dialog = QFileDialog(self)
        dialog.setFileMode(QFileDialog.ExistingFile)
        dialog.setDirectory(regime_path)
        dialog.setNameFilter("Simulation Regime files (*.sreg)")

        if dialog.exec_():
            file = dialog.selectedFiles()[0]
            self.load_regimes_from_file(file)

    def load_regimes_from_file(self, file_name):
        """
        load simulation regime from file
        :param file_name:
        """
        self.regime_file_name = os.path.split(file_name)[-1][:-5]
        self._logger.info("loading regime file: {0}".format(self.regime_file_name))
        with open(file_name.encode(), "r") as f:
            self._regimes += yaml.full_load(f)

        self._update_regime_list()

        if self._regimes:
            self.actSimulateAll.setDisabled(False)

        self._logger.info("loaded {} regimes".format(len(self._regimes)))
        self.statusBar().showMessage("loaded {} regimes.".format(len(self._regimes)), 1000)
        return

    def _update_regime_list(self):
        self.regime_list.clear()
        for reg in self._regimes:
            self._logger.info("adding '{}' to regime list".format(reg["Name"]))
            self.regime_list.addItem(reg["Name"])

    def remove_regime_items(self):
        if self.regime_list.currentRow() >= 0:
            # flag all selected files as invalid
            items = self.regime_list.selectedItems()
            for item in items:
                del self._regimes[self.regime_list.row(item)]
                self.regime_list.takeItem(self.regime_list.row(item))

    @pyqtSlot(QListWidgetItem)
    def regime_dclicked(self, item):
        """
        Apply the selected regime to the current target.
        """
        success = self._apply_regime_by_idx(self.regime_list.row(item))

        self.setQListItemBold(self.regime_list, item, success)
        self.setQListItemBold(self.lastSimList, item, success)
        self.dataPointListWidget.clear()

    def apply_regime_by_name(self, regime_name):
        """
        Apply the regime given by `regime_name` und update the regime index.

        Returns:
            bool: `True` if successful, `False` if errors occurred.
        """
        # get regime idx
        try:
            idx = list(map(itemgetter("Name"), self._regimes)).index(regime_name)
        except ValueError as e:
            self._logger.error("apply_regime_by_name(): Error no regime called "
                               "'{0}'".format(regime_name))
            return False

        # apply
        return self._apply_regime_by_idx(idx)

    def _apply_regime_by_idx(self, index=0):
        """
        Apply the given regime.

        Args:
            index(int): Index of the regime in the `RegimeList` .

        Returns:
            bool: `True` if successful, `False` if errors occurred.
        """
        if index >= len(self._regimes):
            self._logger.error("Invalid index: '{}'".format(index))
            return False

        reg_name = self._regimes[index]["Name"]
        self._logger.info("applying regime '{}'".format(reg_name))

        self._current_regime_index = index
        self._current_regime_name = reg_name

        ret = self.sim.set_regime(self._regimes[index])
        if ret:
            self.statusBar().showMessage("regime {} applied.".format(reg_name),
                                         1000)
        return ret

    @pyqtSlot()
    def start_regime_execution(self):
        """
        Simulate all regimes in the regime list.
        """
        self.actSimulateAll.setText("Stop Simulating &All Regimes")
        self.actSimulateAll.setIcon(QIcon(get_resource("stop_batch.png")))
        self.actSimulateAll.triggered.disconnect(self.start_regime_execution)
        self.actSimulateAll.triggered.connect(self.stop_regime_execution)

        self.runningBatch = True
        self._current_regime_index = -1
        self.regimeFinished.emit()

    def run_next_regime(self):
        """
        Execute the next regime in the regime batch.
        """
        # are we finished?
        if self._current_regime_index == len(self._regimes) - 1:
            self.finishedRegimeBatch.emit(True)
            return

        suc = self._apply_regime_by_idx(self._current_regime_index + 1)
        if not suc:
            self.finishedRegimeBatch.emit(False)
            return

        self.start_simulation()

    @pyqtSlot()
    def stop_regime_execution(self):
        """ Stop the batch process.
        """
        self.stopSimulation.emit()
        self.finishedRegimeBatch.emit(False)

    def regime_batch_finished(self, status):
        self.runningBatch = False
        self.actSimulateAll.setDisabled(False)
        self.actSave.setDisabled(True)

        self.actSimulateAll.setText("Simulate &All Regimes")
        self.actSimulateAll.setIcon(QIcon(get_resource("execute_regimes.png")))
        self.actSimulateAll.triggered.disconnect(self.stop_regime_execution)
        self.actSimulateAll.triggered.connect(self.start_regime_execution)

        if status:
            self.statusLabel.setText("All regimes have been simulated")
            self._logger.info("All Regimes have been simulated")
        else:
            self._logger.error("Batch simulation has been aborted")

        if self._settings.value("control/exit_on_batch_completion") == "True":
            self._logger.info("Shutting down SimulationGUI")
            self.close()

    @pyqtSlot(str, dict, name="new_simulation_data")
    def new_simulation_data(self, status, data):
        """
        Slot to be called when the simulation interface has completed the
        current job and new data is available.

        Args:
            status (str): Status of the simulation, either
                - `finished` : Simulation has been finished successfully or
                - `failed` : Simulation has failed.
            data (dict): Dictionary, holding the simulation data.
        """
        self._logger.info("Simulation {}".format(status))
        self.statusLabel.setText("Simulation {}".format(status))

        self.actSimulateCurrent.setText("&Simulate Current Regime")
        self.actSimulateCurrent.setIcon(QIcon(get_resource("simulate.png")))
        self.actSimulateCurrent.triggered.disconnect(self.stop_simulation)
        self.actSimulateCurrent.triggered.connect(self.start_simulation)

        self.actPlayPause.setDisabled(False)
        self.actStop.setDisabled(False)
        self.actSave.setDisabled(False)
        self.speedControl.setDisabled(False)
        self.timeSlider.setDisabled(False)

        self.sim.simulationProgressChanged.disconnect(self.guiProgress.setValue)
        self.statusBar().removeWidget(self.guiProgress)

        self.stop_animation()

        if data:
            # import new data
            self.currentDataset = data
            self._read_results()
            self._update_data_list()
            self._update_plots()
            self.visualizer.update_config(data["modules"])

            # add results to history
            lastSimCount = self.lastSimList.count()
            lastSimData = {'modules': data['modules'],
                           'results': data['results'],
                           'simulation': data['simulation'],
                           'name': self._current_regime_name,
                           }
            display_name = "{}:{}".format(lastSimCount,
                                          self._current_regime_name)
            self._lastSimulations.append(lastSimData)
            new_item = QListWidgetItem(display_name)
            self.lastSimList.addItem(new_item)
            self.lastSimList.scrollToItem(new_item)
            self.setQListItemBold(self.lastSimList, new_item)

        if self._settings.value("control/autoplay_animation") == "True":
            self.actPlayPause.trigger()

        if self.runningBatch:
            regime_name = self._regimes[self._current_regime_index]["Name"]
            file_name = self._simfile_name(regime_name)
            self._save_data(os.path.join(
                self._settings.value("path/simulation_results"),
                file_name))
            self.regimeFinished.emit()
        else:
            self.actSimulateAll.setDisabled(False)

    def _read_results(self):
        state = self.currentDataset["results"]["Solver"]
        self.interpolator = interp1d(self.currentDataset["results"]["time"],
                                     state,
                                     axis=0,
                                     bounds_error=False,
                                     fill_value=(state[0], state[-1]))
        self.currentStepSize = 1.0 / self.currentDataset["simulation"][
            "measure rate"]
        self.currentStartTime = self.currentDataset["simulation"]["start time"]
        self.currentEndTime = self.currentDataset["simulation"]["end time"]

    def increment_playback_speed(self):
        self.speedControl.setValue(self.speedControl.value()
                                   + self.speedControl.singleStep())

    def decrement_playback_speed(self):
        self.speedControl.setValue(self.speedControl.value()
                                   - self.speedControl.singleStep())

    def reset_playback_speed(self):
        self.speedControl.setValue((self.speedControl.maximum()
                                    - self.speedControl.minimum()) / 2)

    def set_slowest_playback_speed(self):
        self.speedControl.setValue(self.speedControl.minimum())

    def set_fastest_playback_speed(self):
        self.speedControl.setValue(self.speedControl.maximum())

    def update_playback_speed(self, val):
        maximum = self.speedControl.maximum()
        self.playbackGain = 10 ** (3.0 * (val - maximum / 2) / maximum)

    @pyqtSlot()
    def increment_playback_time(self):
        """
        go one time step forward in playback
        """
        if self.playbackTime >= self.currentEndTime:
            self.pause_animation()
            return

        increment = self.playbackGain * self.playbackTimeout / 1000
        self.playbackTime = min(self.currentEndTime,
                                self.playbackTime + increment)
        self.playbackTimeChanged.emit()

    def _update_time_slider(self):
        time_delta = self.currentEndTime - self.currentStartTime
        if time_delta == 0:
            # no meaningful context to update the slider
            return

        rel_pos = (self.playbackTime - self.currentStartTime) / (
            time_delta)
        pos = int(rel_pos * self.timeSliderRange)
        self.timeSlider.blockSignals(True)
        self.timeSlider.setValue(pos)
        self.timeSlider.blockSignals(False)

    def update_playback_time(self):
        """
        adjust playback time to slider value
        """
        self.playbackTime = self.timeSlider.value() / self.timeSliderRange * (
                self.currentEndTime
                - self.currentStartTime
        ) + self.currentStartTime
        self.playbackTimeChanged.emit()

    def update_gui(self):
        """
        Updates the graphical user interface to reflect changes of the current
        display time. 
        
        This includes:
            - timestamp
            - visualisation window
            - time cursors in diagrams
        """
        self.timeLabel.setText("t={0:.3e}".format(self.playbackTime))

        # update timing elements
        self._update_time_slider()
        self._update_time_cursors()

        # update state of rendering
        if self.visualizer is not None and self.interpolator is not None:
            state = self.interpolator(self.playbackTime)
            self.visualizer.update_scene(state)
            if isinstance(self.visualizer, MplVisualizer):
                pass
            elif isinstance(self.visualizer, VtkVisualizer):
                self.vtkWidget.GetRenderWindow().Render()

    def _update_data_list(self):
        # self.dataList.clear()
        self.dataPointListWidget.clear()
        # TODO lets open and check if possible to plot
        # TODO create trees with children instead of plain suffixes
        for module_name, results in self.currentDataset["results"].items():
            if not isinstance(results, np.ndarray):
                continue
            if len(results.shape) == 1:
                self.dataPointListWidget.addItem(module_name)
            elif len(results.shape) == 2:
                for col in range(results.shape[1]):
                    self.dataPointListWidget.addItem(
                        self._build_entry_name(module_name, (col,))
                    )
            elif len(results.shape) == 3:
                for col in range(results.shape[1]):
                    for der in range(results.shape[2]):
                        self.dataPointListWidget.addItem(
                            self._build_entry_name(module_name, (col, der))
                        )

    def _build_entry_name(self, module_name, idx):
        """
        Construct an identifier for a given entry of a module.
        Args:
            module_name (str): name of the module the entry belongs to.
            idx (tuple): Index of the entry.

        Returns:
            str: Identifier to use for display.
        """
        # save the user from defining 1d entries via tuples
        if len(idx) == 1:
            m_idx = idx[0]
        else:
            m_idx = idx

        mod_settings = self.currentDataset["modules"]
        info = mod_settings.get(module_name, {}).get("output_info", None)
        if info:
            if m_idx in info:
                return ".".join([module_name, info[m_idx]["Name"]])

        return ".".join([module_name] + [str(i) for i in idx])

    def _get_index_from_suffix(self, module_name, suffix):
        info = self.currentDataset["modules"].get(module_name, {}).get(
            "output_info", None)
        idx = next((i for i in info if info[i]["Name"] == suffix), None)
        return idx

    def _get_units(self, entry):
        """
        Return the unit that corresponds to a given entry.

        If no information is available, None is returned.

        Args:
            entry (str): Name of the entry. This can either be "Model.a.b" where
                a and b are numbers or if information is available "Model.Signal"
                where signal is the name of that part.

        Returns:

        """
        args = entry.split(".")
        module_name = args.pop(0)
        info = self.currentDataset["modules"].get(module_name, {}).get(
            "output_info", None)
        if info is None:
            return None

        if len(args) == 1:
            try:
                idx = int(args[0])
            except ValueError:
                idx = next((i for i in info if info[i]["Name"] == args[0]),
                           None)
        else:
            idx = (int(a) for a in args)

        return info[idx]["Unit"]

    def _update_plot(self, item):
        # collect data
        if self.currentDataset is None:
            return

        title = item.text(0)
        t = self.currentDataset["results"]["time"]
        docks = self.find_all_plot_docks()
        target = next((d for d in docks if d.title() == title), None)
        if target is None:
            return

        for widget in target.widgets:
            child_names = [item.child(c_idx).text(1)
                           for c_idx in range(item.childCount())]
            del_list = []
            cnt = 0
            for _item in widget.getPlotItem().items:
                if isinstance(_item, pg.PlotDataItem):
                    if _item.name() in child_names:
                        y_data = self._get_data_by_name(_item.name())
                        if y_data is not None:
                            _item.setData(x=t, y=y_data)
                            _item.setPen(self._get_pen(cnt))
                            cnt += 1
                        else:
                            _item.clear()
                    else:
                        del_list.append(_item)

            for _item in del_list:
                widget.getPlotItem().removeItem(_item)

    def plot_data_vector(self, item):
        """
        Creates a plot widget based on the given item.

        If a plot for this item is already open no new plot is created but the
        existing one is raised up again.

        Args:
            item(Qt.ListItem): Item to plot.
        """
        if self.currentDataset is None:
            return

        title = str(item.text(0))

        # create plot widget
        widget = pg.PlotWidget()
        widget.showGrid(True, True)
        widget.getPlotItem().getAxis("bottom").setLabel(text="Time", units="s")

        for idx in range(item.childCount()):
            self.plot_data_vector_member(item.child(idx), widget)

        # add a time line
        time_line = pg.InfiniteLine(self.playbackTime,
                                    angle=90,
                                    movable=False,
                                    pen=pg.mkPen("#FF0000", width=2.0))
        widget.getPlotItem().addItem(time_line)

        coord_item = pg.TextItem(text='', anchor=(0, 1))
        widget.getPlotItem().addItem(coord_item, ignoreBounds=True)

        def _info_wrapper(pos):
            self.update_coord_info(pos, widget, coord_item)

        widget.scene().sigMouseMoved.connect(_info_wrapper)

        # add custom export entries
        widget.scene().contextMenu = [
            QAction("Export as ...", self),
        ]

        def _export_wrapper(export_func):
            def _wrapper():
                return export_func(widget.getPlotItem(),
                                   title,
                                   coord_item,
                                   time_line)
            return _wrapper

        widget.scene().contextMenu[0].triggered.connect(
            _export_wrapper(self.export_plot_item))

        # create dock container and add it to dock area
        dock = pg.dockarea.Dock(title, closable=True)
        dock.addWidget(widget)

        plotWidgets = self.find_all_plot_docks()
        if plotWidgets:
            self.area.addDock(dock, "above", plotWidgets[0])
        else:
            self.area.addDock(dock, "bottom", self.animationDock)

    def update_coord_info(self, pos, widget, coord_item):
        mouse_coords = widget.getPlotItem().vb.mapSceneToView(pos)
        coord_item.setPos(mouse_coords.x(), mouse_coords.y())
        coord_text = "x={:.3e} y={:.3e}".format(mouse_coords.x(),
                                                mouse_coords.y())
        self.coordLabel.setText(coord_text)

        show_info = self._settings.value("view/show_coordinates") == "True"
        if widget.sceneBoundingRect().contains(pos) and show_info:
            coord_item.setText(coord_text.replace(" ", "\n"))
            coord_item.show()
        else:
            coord_item.hide()

    def plot_data_vector_member(self, item, widget):
        idx = item.parent().indexOfChild(item)

        data_name = item.text(1)
        t = self.currentDataset["results"]["time"]
        data = self._get_data_by_name(data_name)
        if data is None:
            t = None

        widget.plot(x=t,
                    y=data,
                    pen=self._get_pen(idx),
                    name=data_name)

    def find_all_plot_docks(self):
        list = []
        for title, dock in self.area.findAll()[1].items():
            if title in self.non_plotting_docks:
                continue
            else:
                list.append(dock)

        return list

    def export_plot_item(self, plot_item, name, coord_item, time_item):
        plot_data = {}
        for i, c in enumerate(plot_item.curves):
            if c.getData() is None:
                continue
            if len(c.getData()) > 2:
                self._logger.warning("Can not handle data format of entry"
                                     "'{}'!".format(c.name()))
                continue
            if "time" not in plot_data:
                plot_data["time"] = c.getData()[0]

            c_name = c.name()
            c_unit = self._get_units(c_name)
            if c_unit is not None:
                c_name += " ({})".format(c_unit)

            plot_data[c_name] = c.getData()[1]

        self.export(plot_data)

    def export(self, plot_data):
        try:
            exporter = Exporter(data_points=plot_data)
        except Exception as e:
            self._logger.error("Can't instantiate exporter! " + str(e))
            return

        last_path = self._settings.value("path/previous_plot_export")
        last_format = self._settings.value("path/previous_plot_format")
        export_formats = ["CSV Data (*.csv)", "PNG Image (*.png)"]
        if last_format == ".png":
            export_formats[:] = export_formats[::-1]
        format_str = ";;".join(export_formats)
        default_file = os.path.join(last_path, "export" + last_format)
        filename = QFileDialog.getSaveFileName(self,
                                               "Export as ...",
                                               default_file,
                                               format_str)

        if filename[0]:
            file, ext = os.path.splitext(filename[0])
            self._settings.setValue("path/previous_plot_export",
                                    os.path.dirname(file))
            if ext == '.csv':
                exporter.export_csv(filename[0])
                self._settings.setValue("path/previous_plot_format", ".csv")
            elif ext == '.png':
                exporter.export_png(filename[0])
                self._settings.setValue("path/previous_plot_format", ".png")
            else:
                self._logger.error("Wrong extension used!")
                return
            self._logger.info("Export successful as '{}'.".format(filename[0]))

    def _get_data_by_name(self, name):
        tmp = name.split(".")
        module_name = tmp[0]
        try:
            raw_data = self.currentDataset["results"][module_name]
        except KeyError:
            return None

        if len(tmp) == 1:
            data = np.array(raw_data)
        elif len(tmp) == 2:
            try:
                idx = int(tmp[1])
            except ValueError:
                idx = self._get_index_from_suffix(module_name, tmp[1])
            if raw_data.ndim != 2 or raw_data.shape[1] <= idx:
                return None
            data = raw_data[:, idx]
        elif len(tmp) == 3:
            try:
                idx = int(tmp[1])
                der = int(tmp[2])
            except ValueError:
                return None
            if raw_data.ndim != 3 or raw_data.shape[1] <= idx \
                    or raw_data.shape[2] <= der:
                return None
            data = raw_data[:, idx, der]
        else:
            raise ValueError("Format not supported")

        return data

    def _update_time_cursors(self):
        """
        Update the time lines of all plot windows
        """
        for title, dock in self.area.findAll()[1].items():
            if title in self.non_plotting_docks:
                continue
            for widget in dock.widgets:
                for item in widget.getPlotItem().items:
                    if isinstance(item, pg.InfiniteLine):
                        item.setValue(self.playbackTime)

    def _update_plots(self):
        """
        Update the data in all plot windows
        """
        root = self.dataPointTreeWidget.invisibleRootItem()
        for i in range(root.childCount()):
            self._update_plot(root.child(i))

    @pyqtSlot(QModelIndex)
    def target_view_changed(self, index):
        self.targetView.resizeColumnToContents(0)

    def postprocessing_clicked(self):
        """
        starts the post- and metaprocessing application
        """
        self._logger.info("launching postprocessor")
        self.statusBar().showMessage("launching postprocessor", 1000)
        if self.postprocessor is None:
            self.postprocessor = PostProcessor(self._settings)

        self.postprocessor.show()

    def reset_camera_clicked(self):
        """
        reset camera in vtk window
        """
        self.visualizer.reset_camera()
        self.vtkWidget.GetRenderWindow().Render()

    def show_info(self):
        icon_lic = open(get_resource("license.txt"), "r").read()
        text = "This application was build using PyMoskito ver. {} .<br />" \
               "PyMoskito is free software distributed under GPLv3. <br />" \
               "It is developed by members of the " \
               "<a href=\'https://tu-dresden.de/ing/elektrotechnik/rst'>" \
               "Institute of Control Theory</a>" \
               " at the <a href=\'https://tu-dresden.de'>" \
               "Dresden University of Technology</a>. <br />" \
               "".format(pkg_resources.require("PyMoskito")[0].version) \
               + "<br />" + icon_lic
        box = QMessageBox.about(self, "PyMoskito", text)

    def show_online_docs(self):
        webbrowser.open("https://pymoskito.readthedocs.org")

    def closeEvent(self, QCloseEvent):
        self._logger.info("Close Event received, shutting down.")
        logging.getLogger().removeHandler(self.textLogger)
        super().closeEvent(QCloseEvent)

    def loadStandardDockState(self):
        for docks in self.find_all_plot_docks():
            docks.close()
        self.area.restoreState(self.standardDockState)

    def setQListItemBold(self, q_list=None, item=None, state=True):
        for i in range(q_list.count()):
            new_font = q_list.item(i).font()
            if q_list.item(i) == item and state:
                new_font.setBold(1)
            else:
                new_font.setBold(0)
            q_list.item(i).setFont(new_font)
        q_list.repaint()



import logging
from abc import ABCMeta, abstractmethod

from matplotlib.backends.backend_qt5agg import (
    NavigationToolbar2QT, FigureCanvasQTAgg as FigureCanvas)
from matplotlib.figure import Figure

__all__ = ["VtkVisualizer", "MplVisualizer"]

try:
    import vtk
    from vtk.qt.QVTKRenderWindowInteractor import *

    class QVTKRenderWindowInteractor(QVTKRenderWindowInteractor):
        """
        overload class to patch problem with zooming in vtk window
        the reason is that the QWheelEvent in PyQt5 hasn't the function delta()
        so we have to replace that with angleDelta()
        the error is caused by vtk 7.0.0
        """
        # override function
        def wheelEvent(self, ev):
            if ev.angleDelta().y() >= 0:
                self._Iren.MouseWheelForwardEvent()
            else:
                self._Iren.MouseWheelBackwardEvent()

except ImportError as e:
    QVTKRenderWindowInteractor = None


class Visualizer(metaclass=ABCMeta):
    """
    Base Class for animation
    """

    def __init__(self):
        self.can_reset_view = False
        self._config = {}
        self._logger = logging.getLogger(self.__class__.__name__)

    def update_config(self, config):
        """
        Hook to update the current visualization configuration
        :param config: Current simulation config
        """
        self._config = config

    @abstractmethod
    def update_scene(self, x):
        """
        Hook to update the current visualization state
        :param x: system state vector
        """
        pass


class DummyVisualizer(Visualizer):

    def update_scene(self, x):
        pass


class VtkVisualizer(Visualizer):
    """
    Base Class with some functionality the help visualizing the system using vtk
    """

    def __init__(self, renderer):
        Visualizer.__init__(self)

        assert isinstance(renderer, vtk.vtkRenderer)
        self.ren = renderer

        self.can_reset_view = False
        self.position = None
        self.focal_point = None
        self.view_up = None
        self.view_angle = None
        self.parallel_projection = None
        self.parallel_scale = None
        self.clipping_range = None

    def reset_camera(self):
        """
        Reset camera to original view.
         
        Will be available if you implement the attributes below and set the
        'can_reset_view' flag.
        """
        if self.can_reset_view:
            camera = self.ren.GetActiveCamera()
            camera.SetPosition(self.position)
            camera.SetFocalPoint(self.focal_point)
            camera.SetViewUp(self.view_up)
            camera.SetViewAngle(self.view_angle)
            camera.SetParallelProjection(self.parallel_projection)
            camera.SetParallelScale(self.parallel_scale)
            camera.SetClippingRange(self.clipping_range)
        else:
            self.ren.ResetCamera()

    def save_camera_pose(self):
        # add camera reset functionality
        camera = self.ren.GetActiveCamera()
        self.position = camera.GetPosition()
        self.focal_point = camera.GetFocalPoint()
        self.view_up = camera.GetViewUp()
        self.view_angle = camera.GetViewAngle()
        self.parallel_projection = camera.GetParallelProjection()
        self.parallel_scale = camera.GetParallelScale()
        self.clipping_range = camera.GetClippingRange()

        self.can_reset_view = True


class MplVisualizer(Visualizer):
    """
    Base Class with some function the help visualizing the system using
    matplotlib
    """

    def __init__(self, q_widget, q_layout):
        Visualizer.__init__(self)
        self.qWidget = q_widget
        self.qLayout = q_layout
        self.dpi = 100
        self.fig = Figure((5.0, 4.0), facecolor='white', dpi=self.dpi)
        self.canvas = FigureCanvas(self.fig)
        self.canvas.setParent(self.qWidget)
        self.axes = self.fig.add_subplot(111)
        self.qLayout.addWidget(self.canvas)
        self.qWidget.setLayout(self.qLayout)



import logging
import matplotlib as mpl

from .tools import get_figure_size

_logger = logging.getLogger("mpl_settings")

orig_settings = {**mpl.rcParams}
latex_settings = {
    # change this if using contex, xetex or lualatex
    "pgf.texsystem": "pdflatex",
    # use LaTeX to write all text
    "text.usetex": True,
    'font.family': 'lmodern',
    # blank entries should cause plots to inherit fonts from the document
    # "font.serif": [],
    # "font.sans-serif": [],
    # "font.monospace": [],
    # "text.fontsize": 11,
    "legend.fontsize": 9,  # Make the legend/label fonts a little smaller
    "xtick.labelsize": 9,
    "ytick.labelsize": 9,
    "figure.figsize": get_figure_size(1),  # default fig size of 1\textwidth
    "lines.linewidth": 0.5,
    "axes.labelsize": 11,  # LaTeX default is 10pt font.
    "axes.linewidth": 0.5,
    "axes.unicode_minus": False,
    # subfig related
    "figure.subplot.left": 0.1,
    "figure.subplot.right": 0.95,
    "figure.subplot.bottom": 0.125,
    "figure.subplot.top": 0.95,
    # the amount of width reserved for blank space between subplots
    "figure.subplot.wspace": 0.4,
    # the amount of height reserved for white space between subplots
    "figure.subplot.hspace": 0.4,
    # Patches are graphical objects that fill 2D space, like polygons or circles
    "patch.linewidth": 0.5,
}


def enable_latex():
    _logger.info("LaTeX export enabled")
    mpl.rcParams['text.latex.preamble'] += r"\usepackage{lmodern}"
    mpl.rcParams['text.latex.preamble'] += r"\usepackage{siunitx}"
    mpl.rcParams.update(latex_settings)


def disable_latex():
    _logger.info("LaTeX export disabled")
    mpl.rcParams = orig_settings



# -*- coding: utf-8 -*-
import sys
import os
import importlib
from PyQt5.QtWidgets import QApplication

import pymoskito as pm

if __name__ == '__main__':
    # This part lets us execute this example with:
    #  $python -m pymoskito.examples.name
    pkg_path = os.path.dirname(os.path.abspath(__file__))
    pkg_name = pkg_path.split(os.path.sep)[-1]
    parent_dir = os.path.dirname(pkg_path)
    sys.path.insert(0, parent_dir)

    if __package__ is None or __package__ == '':
        importlib.import_module(pkg_name)
        __package__ = pkg_name

    # create an Application instance (needed)
    app = QApplication([])
    prog = None
    sim = pm.SimulationGui()

    # load defaults
    sim.load_regimes_from_file(os.path.join(parent_dir,
                                            pkg_name,
                                            "default.sreg"))
    sim.apply_regime_by_name("test")
    sim.show()

    app.exec_()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_modules
----------------------------------

Tests for simulation modules.
"""

import unittest
from collections import OrderedDict
import pickle
import os

import scipy.signal as sig
import numpy as np

import pymoskito as pm
from pymoskito.simulation_modules import (
    SimulationModule, Model
)


class DummyModule(SimulationModule):
    """
    Possibly the dumbest implementation of a Simulation Module

    The attribute 'public_settings' is implemented as ordered dict to have
    control over the order in which the entries will appear in the settings
    window.
    """
    public_settings = OrderedDict()

    def calc_output(self, input_vector):
        """
        This method has to be implemented, returning the input should do the
        job.
        """
        return "bazz"


class TestSimulationModule(unittest.TestCase):
    def setUp(self):
        pass

    def test_init(self):
        # Type Error due to abstract base class
        with self.assertRaises(TypeError):
            SimulationModule()

        # Initialization has te be done with correct settings
        with self.assertRaises(AssertionError):
            DummyModule(settings=None)

        # This is the easiest way
        settings = DummyModule.public_settings

        # check default attributes
        m = DummyModule(settings)
        m._settings["tick divider"] = 1
        m._settings["step width"] = None

        # modules entry has to be deleted after init is completed
        with self.assertRaises(KeyError):
            m._settings["modules"]

        # check acquisition of attributes
        settings["tick divider"] = 1234
        settings["step width"] = 13.26  # this _CANNOT_ be known a priori
        info = {0: dict(Name="Centripetal Force",
                        Unit="Newton [N]")}
        settings["output info"] = info
        m = DummyModule(settings)
        self.assertEqual(m._settings["tick divider"], 1234)
        self.assertEqual(m._settings["step width"], None)
        self.assertEqual(m._settings["output info"], info)

    def test_properties(self):
        settings = DummyModule.public_settings
        settings["tick divider"] = 1234
        settings["step width"] = 13.26  # this _CANNOT_ be known a priori
        m = DummyModule(settings)

        self.assertTrue(isinstance(m.public_settings, OrderedDict))

        self.assertEqual(m.tick_divider, 1234)
        self.assertEqual(m.step_width, None)

        m.step_width = 13.37
        self.assertEqual(m.step_width, 13.37)

    def test_calc_output(self):
        settings = DummyModule.public_settings
        settings["tick divider"] = 1234
        settings["step width"] = 13.26  # this _CANNOT_ be known a priori
        m = DummyModule(settings)

        input_data = dict(foo="bar")  # later dict of pairs module-type: output
        out = m.calc_output(input_data)


class TrajectoryTestCase(unittest.TestCase):

    def setUp(self):
        self.d_order = 7

    def test_setpoint(self):
        # check defaults
        st = pm.Setpoint.public_settings
        self.assertEqual(st["Setpoint"], [0])

        st["differential_order"] = self.d_order
        st["Setpoint"] = [1, 4]
        tr = pm.Setpoint(st)

        # check properties
        self.assertEqual(tr._settings["differential_order"], self.d_order)
        self.assertEqual(tr._settings["Setpoint"], [1, 4])

        # check output
        out_1 = tr.calc_output(dict(time=1))
        ref_out = np.zeros((2, self.d_order + 1))
        ref_out[0, 0] = 1
        ref_out[1, 0] = 4
        np.testing.assert_array_equal(out_1, ref_out)

        # setpoint should be time independent
        out_2 = tr.calc_output(dict(time=10))
        np.testing.assert_array_equal(out_1, out_2)

    def test_smooth_transition(self):
        # check defaults
        st = pm.SmoothTransition.public_settings
        self.assertEqual(st["states"], [[0, 1]])
        self.assertEqual(st["start time"], 0)
        self.assertEqual(st["delta t"], 5)

        st["states"] = [-4, 9]
        st["start time"] = 10
        st["delta t"] = 7
        st["differential_order"] = self.d_order
        # test deprecation
        with self.assertWarns(DeprecationWarning):
            tr = pm.SmoothTransition(st)

        st["states"] = [[-4, 9], [10, 20]]
        tr = pm.SmoothTransition(st)

        # check properties
        np.testing.assert_array_equal(tr._settings["states"], st["states"])
        self.assertEqual(tr._settings["start time"], st["start time"])
        self.assertEqual(tr._settings["delta t"], st["delta t"])
        self.assertEqual(tr._settings["differential_order"], self.d_order)

        # check trivial outputs in the beginning
        ref_out = np.zeros((2, self.d_order + 1))

        # check trivial outputs in the beginning
        ref_out[0, 0] = -4
        ref_out[1, 0] = 10
        times = np.linspace(0, 10)
        for t in times:
            out = tr.calc_output(dict(time=t))
            np.testing.assert_array_equal(out, ref_out)

        # check trivial outputs in the end
        ref_out[0, 0] = 9
        ref_out[1, 0] = 20
        times = np.linspace(17, 100)
        for t in times:
            out = tr.calc_output(dict(time=t))
            np.testing.assert_array_equal(out, ref_out)


class DummyModel(Model):

    public_settings = OrderedDict([("initial state", [1, 2, 3, 4])])

    def __init__(self, settings):
        super().__init__(settings)

    def state_function(self, t, x, args):
        return np.zeros((4,))

    def calc_output(self, input_vector):
        return 0


class TestModel(unittest.TestCase):

    def test_init(self):
        settings = OrderedDict([("initial state", [1, 2, 3, 4])])

        # input_count and state_count are missing
        with self.assertRaises(AssertionError):
            DummyModel(settings)

        settings["input_count"] = 1
        settings["state_count"] = 3

        # state count oes not match initial state
        with self.assertRaises(AssertionError):
            DummyModel(settings)

        settings["state_count"] = 4
        m = DummyModel(settings)

    def test_properties(self):
        settings = OrderedDict([("initial state", [1, 2, 3, 4])])
        settings["input_count"] = 1
        settings["state_count"] = 4
        m = DummyModel(settings)

        self.assertEqual(m.initial_state, [1, 2, 3, 4])


class StateSpaceModulesTest(unittest.TestCase):

    def setUp(self):
        # generate StateSpace model
        self.sys = sig.TransferFunction([5], [10, 1]).to_ss()

        # generate extra data
        self.ss_inputs = [1]
        self.ss_outputs = [4]

        data = dict(system=self.sys,
                    op_inputs=self.ss_inputs,
                    op_outputs=self.ss_outputs
                    )

        self.config_path = os.sep.join([os.path.dirname(__file__),
                                       "test_sys.pkl"])
        if not os.path.exists(self.config_path):
            with open(self.config_path, "wb") as f:
                pickle.dump(data, f)

    def test_model_init(self):
        settings = pm.LinearStateSpaceModel.public_settings

        # path should point to a real file
        settings["config file"] = "A: [2, 3], B = 1"
        with self.assertRaises(AssertionError):
            pm.LinearStateSpaceModel(settings)

        # neither initial state nor initial output present
        settings["config file"] = self.config_path
        with self.assertRaises(ValueError):
            pm.LinearStateSpaceModel(settings)

        x0 = np.array([12])
        settings["initial state"] = x0
        m = pm.LinearStateSpaceModel(settings)
        self.assertEqual(m.initial_state, x0)

        settings["initial state"] = None
        settings["initial output"] = self.sys.C @ x0 + self.ss_outputs
        m = pm.LinearStateSpaceModel(settings)
        self.assertEqual(m.initial_state, x0)

    def test_controller_init(self):
        """ Same thing goes for the statespace controller
        """
        settings = pm.LinearStateSpaceController.public_settings

        # path should point to a real file
        settings["config file"] = "A: [2, 3], B = 1"
        with self.assertRaises(AssertionError):
            pm.LinearStateSpaceController(settings)

        # check if poles are placed as intended
        settings["config file"] = self.config_path
        settings["poles"] = [-5]
        c = pm.LinearStateSpaceController(settings)
        eig_vals = np.linalg.eigvals(c.ss.A - c.ss.B @ c.feedback)
        np.testing.assert_array_almost_equal(settings["poles"], eig_vals)


class ModelInputLimiterTestCase(unittest.TestCase):

    def setUp(self):
        self.scalar_limits = [-5, 5]
        self.scalar_sample = np.linspace(-10, 10)
        self.vectorial_limits = [[-2, 2], [0, None], [-10, None]]
        self.vectorial_sample = np.hstack([self.scalar_sample,
                                          self.scalar_sample * 10,
                                          self.scalar_sample - 10])

    def test_init(self):
        st = pm.ModelInputLimiter.public_settings

        # check default args
        l = pm.ModelInputLimiter(st)
        self.assertEqual(l._settings["input_signal"], "ModelMixer")
        self.assertEqual(l._settings["Limits"], [None, None])

        # check attributes
        np.testing.assert_array_equal(l.limits,
                                      np.atleast_2d([None, None]))

    def test_scalar_case(self):
        # no limits
        l = pm.ModelInputLimiter(pm.ModelInputLimiter.public_settings)
        out = np.array([l.calc_output(dict(ModelMixer=val))
                        for val in self.scalar_sample])
        np.array_equal(out, self.scalar_sample)

        l = pm.ModelInputLimiter(dict(Limits=self.scalar_limits))
        np.testing.assert_array_equal(l.limits,
                                      np.atleast_2d(self.scalar_limits))
        out = np.array([l.calc_output(dict(ModelMixer=val))
                        for val in self.scalar_sample])
        self.assertTrue(out.min() >= self.scalar_limits[0])
        self.assertTrue(out.max() <= self.scalar_limits[1])

    def test_vectorial_case(self):
        l = pm.ModelInputLimiter(dict(Limits=self.vectorial_limits))
        np.testing.assert_array_equal(l.limits,
                                      np.atleast_2d(self.vectorial_limits))

        out = np.array([l.calc_output(dict(ModelMixer=vals))
                        for vals in self.vectorial_sample])
        for idx, col in enumerate(out.T):
            self.assertTrue(col.min() >= self.vectorial_limits[idx][0])
            self.assertTrue(col.max() <= self.vectorial_limits[idx][1])

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from .processing_core import ProcessingModule
from .simulation_modules import SimulationModule, SignalMixer
from .visualization import Visualizer

__all__ = ["register_simulation_module", "register_processing_module",
           "register_visualizer",
           "get_registered_modules",
           "get_registered_simulation_modules",
           "get_registered_processing_modules",
           "get_registered_visualizers",
           ]

"""
wrapper for easy user interaction
"""

_registry = {}


def register_module(module_cls, module_type, cls, type_check=True):
    """
    Main hook to register a class (pymoskito module) in the pymoskito framework.
    
    Args:
        module_cls (cls): Either :py:class:`pm.SimulationModule` or 
            :py:class:`pm.ProcessingModule` 
        module_type (cls): Type of the module, these are either the classes 
            listed in `simulation_modules.py`, 
            :py:class:`pm.PostProcessingModule` or 
            :py:class:`pm.MetaProcessingModule` .
        cls (cls): Class to be registered.
        type_check (bool): Perform type checking (defaults to True)
     
    Raises:
        TypeError: If `cls` is not a subclass of `module_type` .
        ValueError: If `cls` is already registered for `module_type` .

    """
    if type_check:
        if not issubclass(cls, module_type):
            raise TypeError("Module must match Type to be registered for "
                            "{0} <> {1}".format(cls, module_type))

    cls_entry = _registry.get(module_cls, {})
    entry = cls_entry.get(module_type, [])
    increment = (cls, cls.__name__)
    if increment in entry:
        raise ValueError("class {} already registered for module "
                         "{}!".format(cls, module_cls))

    entry.append(increment)

    cls_entry[module_type] = entry
    _registry[module_cls] = cls_entry

    cls_entry[module_type.__name__] = entry
    _registry[module_cls.__name__] = cls_entry


def get_registered_modules(module_cls, module_type):
    """
    Return registered classes of a module type for a specific module class.
    
    Args:
        module_cls (cls): Either :py:class:`pm.SimulationModule` or 
            :py:class:`pm.ProcessingModule` 
        module_type (cls): Type of the module, these are either the classes 
            listed in `simulation_modules.py`, 
            :py:class:`pm.PostProcessingModule` or
            :py:class:`pm.MetaProcessingModule` .
            
    Returns:
        list: List of (obj, string) tuples.
    """
    return _registry.get(module_cls, {}).get(module_type, [])


def get_module_class_by_name(module_cls, module_type, module_name):
    """
    return class object for given name
    :param module_name:
    :param module_type:
    :param module_cls:
    """
    return next((mod[0] for mod in get_registered_modules(module_cls,
                                                          module_type)
                 if mod[1] == module_name), None)


def register_simulation_module(module_type, cls):
    """
    main hook to register a module in the pymoskito framework
    :param module_type:
    :param cls: class to be registered
    :return: None
    """
    if not issubclass(cls, SimulationModule):
        raise TypeError("Only Simulation Modules can be registered!")
    if not issubclass(cls, module_type):
        # more complex inheritance allowed for mixers
        if not (issubclass(module_type, SignalMixer)
                and issubclass(cls, SignalMixer)):
            raise TypeError("Class must be instance of {0}".format(module_type))

    register_module(SimulationModule, module_type, cls, type_check=False)


def get_registered_simulation_modules(module_type):
    """
    main hook to retrieve registered classes for a specific simulation module
    :param module_type:
    :return:
    """
    return get_registered_modules(SimulationModule, module_type)


def get_simulation_module_class_by_name(module_type, module_name):
    """
    Return the class of a certain simulation module given its registered name.
    
    Args:
        module_type (cls): Type of the module, 
            see :py:func:`register_simulation_module` .
        module_name (str): name of the module. 

    Returns:

    """
    return get_module_class_by_name(SimulationModule, module_type, module_name)


def register_processing_module(module_type, cls):
    """
    main hook to register a class in the pymoskito framework
    :param module_type:
    :param cls: class to be registered
    :return: None
    """
    if not issubclass(cls, ProcessingModule):
        raise TypeError("Only PostProcessing Modules can be registered!")

    register_module(ProcessingModule, module_type, cls)


def get_registered_processing_modules(module_type):
    """
    main hook to retrieve registered classes for a specific simulation module
    :param module_type:
    :return:
    """
    return get_registered_modules(ProcessingModule, module_type)


def get_processing_module_class_by_name(module_type, module_name):
    return get_module_class_by_name(ProcessingModule, module_type, module_name)


def register_visualizer(vis_cls):
    """
    hook to register a visualizer for the simulation GUI

    :param vis_cls: class to be registered
    """
    if not issubclass(vis_cls, Visualizer):
        raise TypeError("Module must match type to be registered for! "
                        "{0} <> {1}".format(vis_cls, Visualizer))

    cls_entry = _registry.get(Visualizer, [])
    increment = (vis_cls, vis_cls.__name__)
    if increment in cls_entry:
        raise ValueError("class {0} already registered as visualizer!"
                         "".format(vis_cls))

    cls_entry.append(increment)
    _registry[Visualizer] = cls_entry
    _registry[Visualizer.__name__] = cls_entry


def get_registered_visualizers():
    """
    hook to retrieve registered visualizers
    :return: visualizer class
    """
    return _registry.get(Visualizer, [])



# -*- coding: utf-8 -*-
from collections import OrderedDict

import numpy as np
import sympy as sp
from scipy.integrate import ode
from sympy import sin, cos

import pymoskito as pm
from pymoskito.tools import swap_cols, swap_rows
from . import settings as st
from .linearization import linearise_system


class LuenbergerObserver(pm.Observer):
    """
    Luenberger Observer that uses a simple euler forward integration
    """
    public_settings = OrderedDict([
        ("initial state", [0, 0, 0, 0]),
        ("poles", [-20, -20, -20, -20]),
        ("steady state", [0, 0, 0, 0]),
        ("steady tau", 0),
        ("tick divider", 1),
    ])

    def __init__(self, settings):
        settings.update(output_dim=4)
        super().__init__(settings)

        self.a_mat, self.b_mat, self.c_mat = linearise_system(
            self._settings["steady state"],
            self._settings["steady tau"])
        self.L = pm.place_siso(self.a_mat.T,
                               self.c_mat.T,
                               self._settings["poles"]).T
        self.output = np.array(self._settings["initial state"], dtype=float)

    def state_func(self, t, q, args):
        x_o = q
        u = args[0]
        y = args[1]

        dx_o = ((self.a_mat - self.L @ self.c_mat) @ x_o
                + self.b_mat @ u
                + self.L @ y)
        return dx_o

    def _observe(self, time, system_input, system_output):
        # np.copyto(self.output, self.next_output)
        if system_input is not None:
            dy = self.state_func(time, self.output,
                                 (system_input, system_output))
            self.output += self.step_width * dy
        return self.output


class LuenbergerObserverReduced(pm.Observer):
    """
    Reduced luenberger observer that uses a simple euler forward integration
    
    The observers ode solver has been reduced to EULER forward which enables 
    a better performance.
    The output is calculated in a reduced and transformed system.
    """
    public_settings = OrderedDict([
        ("initial state", [0, 0, 0, 0]),
        ("poles", [-3, -3, -3]),
        ("steady state", [0, 0, 0, 0]),
        ("steady tau", 0),
        ("tick divider", 1),
    ])

    def __init__(self, settings):
        settings.update(output_dim=4)
        super().__init__(settings)

        self.a_mat, self.b_mat, self.c_mat = linearise_system(
            self._settings["steady state"],
            self._settings["steady tau"])

        self.n = self.a_mat.shape[0]
        self.r = self.c_mat.shape[0]

        if self.c_mat.shape != (1, 4):
            raise Exception("LuenbergerObserverReduced only usable for SISO")

        # which cols and rows must be sorted
        self.switch = 0
        for i in range(self.c_mat.shape[1]):
            if self.c_mat[0, i] == 1:
                self.switch = i
                break

        # sort A, B, C for measured and unmeasured states
        self.a_mat = swap_cols(self.a_mat, 0, self.switch)
        self.a_mat = swap_rows(self.a_mat, 0, self.switch)
        self.C = swap_cols(self.c_mat, 0, self.switch)
        self.B = swap_rows(self.b_mat, 0, self.switch)

        # get reduced system
        self.a_mat_11 = self.a_mat[0:self.r, 0:self.r]
        self.a_mat_12 = self.a_mat[0:self.r, self.r:self.n]
        self.a_mat_21 = self.a_mat[self.r:self.n, 0:self.r]
        self.a_mat_22 = self.a_mat[self.r:self.n, self.r:self.n]

        self.b_1 = self.b_mat[0:self.r, :]
        self.b_2 = self.b_mat[self.r:self.n, :]

        self.L = pm.place_siso(self.a_mat_22.T,
                               self.a_mat_12.T,
                               self._settings["poles"]).T
        self.output = np.array(self._settings["initial state"], dtype=float)

    def _observe(self, time, system_input, system_output):
        if system_input is None:
            return self.output

        # transform ((un-)measured) and collect necessary from observer_output
        x_o = self.output
        x_o = swap_rows(x_o, 0, self.switch)
        x_o = x_o[self.r:self.n]

        u = system_input
        y = system_output

        # transform system and eliminate ydot
        x_o_t = x_o - self.L @ y
        dy = ((self.a_mat_22 - self.L @ self.a_mat_12) @ x_o_t
              + (self.a_mat_21 - self.L @ self.a_mat_11
                 + (self.a_mat_22 - self.L @ self.a_mat_12) @ self.L) @ y
              + (self.b_2 - self.L @ self.b_1) @ u
              )

        # EULER integration
        x_o_t += self.step_width * dy

        # transform system back to original observer coordinates
        x_o = x_o_t + self.L @ y

        out = np.hstack((y, x_o))

        # change state order back to the original order
        self.output = swap_rows(out, 0, self.switch)

        return self.output


class HighGainObserver(pm.Observer):
    """
    High Gain Observer for nonlinear systems
    """
    public_settings = OrderedDict([
        ("initial state", [0, 0, 0, 0]),
        ("poles", [-10, -10, -10, -10]),
        ("tick divider", 1),
    ])

    def __init__(self, settings):
        settings.update(output_dim=4)
        super().__init__(settings)

        params = sp.symbols('x1, x2, x3, x4, tau')
        x1, x2, x3, x4, tau = params
        x = [x1, x2, x3, x4]
        h = sp.Matrix([[x1]])
        f = sp.Matrix([[x2],
                       [st.B * x1 * x4 ** 2 - st.B * st.G * sin(x3)],
                       [x4],
                       [(tau - 2 * st.M * x1 * x2 * x4
                         - st.M * st.G * x1 * cos(x3)) / (
                        st.M * x1 ** 2 + st.J + st.Jb)]])

        q = sp.Matrix(pm.lie_derivatives(h, f, x, len(x) - 1))
        dq = q.jacobian(x)

        if dq.rank() != len(x):
            raise Exception("System might not be observable")

        # gets p = [p0, p1, ... pn-1]
        p = pm.char_coefficients(self._settings["poles"])

        k = p[::-1]
        l = dq.inv() @ k

        mat2array = [{'ImmutableMatrix': np.array}, 'numpy']
        self.h_func = sp.lambdify((x1, x2, x3, x4, tau), h, modules=mat2array)
        self.l_func = sp.lambdify((x1, x2, x3, x4, tau), l, modules=mat2array)
        self.f_func = sp.lambdify((x1, x2, x3, x4, tau), f, modules=mat2array)

        self.output = np.array(self._settings["initial state"], dtype=float)

    def _observe(self, time, system_input, system_output):
        if system_input is None:
            return self.output

        y = system_output
        u = system_input

        l_np = self.l_func(*self.output, u)
        f_np = self.f_func(*self.output, u)
        h_x_o = self.h_func(self.output[0], 0, 0, 0, 0)

        dx_o = f_np + l_np @ (y - h_x_o)

        # EULER integration
        self.output += np.squeeze(self.step_width * dx_o, axis=1)

        return self.output


class LuenbergerObserverInt(pm.Observer):
    """
    Luenberger Observer that uses solver to integrate
    """
    public_settings = OrderedDict([
        ("initial state", [0, 0, 0, 0]),
        ("poles", [-3, -3, -3, -3]),
        ("steady state", [0, 0, 0, 0]),
        ("steady tau", 0),
        ("step width", .001),
        ("tick divider", 1),
        ("Method", "adams"),
        ("step size", 1e-2),
        ("rTol", 1e-6),
        ("aTol", 1e-9),
        ("start time", 0),
        ("end time", 5)
    ])

    def __init__(self, settings):
        settings.update(output_dim=4)
        super().__init__(settings)

        self.a_mat, self.b_mat, self.c_mat = linearise_system(
            self._settings["steady state"],
            self._settings["steady tau"])
        self.L = pm.place_siso(self.a_mat.T,
                               self.c_mat.T,
                               self._settings["poles"]).T
        self.step_width = self._settings["tick divider"] * self._settings["step width"]
        self.output = np.array(self._settings["initial state"], dtype=float)

        self.solver = ode(self.linear_state_function)
        self.solver.set_integrator("dopri5",
                                   method=self._settings["Method"],
                                   rtol=self._settings["rTol"],
                                   atol=self._settings["aTol"])
        self.solver.set_initial_value(self._settings["initial state"])
        self.nextObserver_output = self.output

    def linear_state_function(self, t, q, args):
        x1_o, x2_o, x3_o, x4_o = q
        u = args[0]
        y = args[1]

        x_o = q
        dx_o = (self.a_mat - self.L @ self.c_mat) @ x_o + self.b_mat @ u + self.L @ y
        return dx_o

    def _observe(self, time, system_input, system_output):
        if system_input is not None:
            self.solver.set_f_params((system_input, system_output))
            self.output = self.solver.integrate(time)
        
        return self.output

pm.register_simulation_module(pm.Observer, LuenbergerObserver)
pm.register_simulation_module(pm.Observer, LuenbergerObserverInt)
pm.register_simulation_module(pm.Observer, LuenbergerObserverReduced)
pm.register_simulation_module(pm.Observer, HighGainObserver)



# -*- coding: utf-8 -*-
import matplotlib as mpl
import matplotlib.patches

import pymoskito as pm
from . import settings as st


class MplTankVisualizer(pm.MplVisualizer):

    def __init__(self, q_widget, q_layout):
        pm.MplVisualizer.__init__(self, q_widget, q_layout)

        self.rT = 0.06
        self.hT = 0.3
        self.axes.set_xlim(-4 * self.rT, 4 * self.rT + 2 * self.rT + 4 * self.rT)
        self.axes.set_ylim(-0.1, self.hT + 0.1)
        self.axes.set_aspect("equal")
        self.axes.get_xaxis().set_visible(False)
        self.axes.get_yaxis().set_visible(False)

        tube1_house = mpl.patches.Rectangle(xy=[0, 0],
                                            width=2 * self.rT,
                                            height=self.hT,
                                            linewidth=2,
                                            fill=False,
                                            zorder=2)
        self.axes.add_patch(tube1_house)

        self.tube1 = mpl.patches.Rectangle(xy=[0, 0],
                                           width=2 * self.rT,
                                           height=0,
                                           linewidth=0.1,
                                           fill=True,
                                           color='blue',
                                           zorder=1)
        self.axes.add_patch(self.tube1)

        tube2_house = mpl.patches.Rectangle(xy=[4 * self.rT + 2 * self.rT, 0],
                                            width=2 * self.rT,
                                            height=self.hT,
                                            linewidth=2,
                                            fill=False,
                                            zorder=2)
        self.axes.add_patch(tube2_house)

        self.tube2 = mpl.patches.Rectangle(xy=[4 * self.rT + 2 * self.rT, 0],
                                           width=2 * self.rT,
                                           height=0,
                                           linewidth=0.1,
                                           fill=True,
                                           color='blue',
                                           zorder=1)
        self.axes.add_patch(self.tube2)

        conT1_house_above = mpl.lines.Line2D([-3 * self.rT, -3 * self.rT + 3 * self.rT],
                                             [0.1 * self.hT + 0.1 * self.hT, 0.1 * self.hT + 0.1 * self.hT],
                                             linewidth=2,
                                             color='black',
                                             zorder=3)
        self.axes.add_line(conT1_house_above)

        conT1_house_below = mpl.lines.Line2D([-3 * self.rT, -3 * self.rT + 3 * self.rT],
                                             [0.1 * self.hT, 0.1 * self.hT],
                                             linewidth=2,
                                             color='black',
                                             zorder=3)
        self.axes.add_line(conT1_house_below)

        self.conT1 = mpl.patches.Rectangle(xy=[(-3 * self.rT), 0.1 * self.hT],
                                           width=(3 * self.rT),
                                           height=0.1 * self.hT,
                                           linewidth=2,
                                           fill=True,
                                           color='white',
                                           zorder=3)
        self.axes.add_patch(self.conT1)

        conT1T2_house_above = mpl.lines.Line2D([2 * self.rT, 2 * self.rT + 4 * self.rT],
                                               [0.1 * self.hT + 0.1 * self.hT, 0.1 * self.hT + 0.1 * self.hT],
                                               linewidth=2,
                                               color='black',
                                               zorder=3)
        self.axes.add_line(conT1T2_house_above)

        conT1T2_house_below = mpl.lines.Line2D([2 * self.rT, 2 * self.rT + 4 * self.rT],
                                               [0.1 * self.hT, 0.1 * self.hT],
                                               linewidth=2,
                                               color='black',
                                               zorder=3)
        self.axes.add_line(conT1T2_house_below)

        self.conT1T2 = mpl.patches.Rectangle(xy=[(2 * self.rT), 0.1 * self.hT],
                                             width=(4 * self.rT),
                                             height=0.1 * self.hT,
                                             linewidth=2,
                                             fill=True,
                                             color='white',
                                             zorder=3)
        self.axes.add_patch(self.conT1T2)

        conT2_house_above = mpl.lines.Line2D([4 * self.rT + 4 * self.rT, 4 * self.rT + 4 * self.rT + self.rT],
                                             [0.1 * self.hT + 0.1 * self.hT, 0.1 * self.hT + 0.1 * self.hT],
                                             linewidth=2,
                                             color='black',
                                             zorder=3)
        self.axes.add_line(conT2_house_above)

        conT2_house_below = mpl.lines.Line2D([4 * self.rT + 4 * self.rT, 4 * self.rT + 4 * self.rT + self.rT],
                                             [0.1 * self.hT, 0.1 * self.hT],
                                             linewidth=2,
                                             color='black',
                                             zorder=3)
        self.axes.add_line(conT2_house_below)

        self.conT2 = mpl.patches.Rectangle(xy=[(4 * self.rT + 4 * self.rT), 0.1 * self.hT],
                                           width=self.rT,
                                           height=0.1 * self.hT,
                                           linewidth=2,
                                           fill=True,
                                           color='white',
                                           zorder=3)
        self.axes.add_patch(self.conT2)

        pump_house = mpl.patches.Circle(xy=[-1.5 * self.rT, 0.15 * self.hT],
                                        radius=0.1 * self.hT,
                                        linewidth=1,
                                        edgecolor='black',
                                        facecolor='white',
                                        zorder=10)
        self.axes.add_patch(pump_house)

        pump = mpl.patches.Polygon([[-1.5 * self.rT, 0.15 * self.hT + 0.1 * self.hT],
                                    [-1.5 * self.rT + 0.1 * self.hT, 0.15 * self.hT],
                                    [-1.5 * self.rT, 0.15 * self.hT - 0.1 * self.hT]],
                                   linewidth=1,
                                   closed=False,
                                   edgecolor='black',
                                   facecolor='white',
                                   zorder=10)
        self.axes.add_patch(pump)

        ventil_house = mpl.patches.Rectangle(xy=[3.25 * self.rT, 0.05 * self.hT],
                                             width=1.5 * self.rT,
                                             height=0.2 * self.hT,
                                             fill=True,
                                             color='white',
                                             zorder=10)
        self.axes.add_patch(ventil_house)

        ventil = mpl.patches.Polygon([[3.25 * self.rT, 0.05 * self.hT],
                                      [3.25 * self.rT + 1.5 * self.rT, 0.05 * self.hT + 0.2 * self.hT],
                                      [3.25 * self.rT + 1.5 * self.rT, 0.05 * self.hT],
                                      [3.25 * self.rT, 0.05 * self.hT + 0.2 * self.hT]],
                                     linewidth=1,
                                     edgecolor='black',
                                     facecolor='white',
                                     zorder=10)
        self.axes.add_patch(ventil)

        self.axes.text(self.rT, self.hT + 0.01, 'Tank 1', horizontalalignment='center')
        self.axes.text(4 * self.rT + 3 * self.rT, self.hT + 0.01, 'Tank 2', horizontalalignment='center')

    def update_scene(self, x):
        if 0 < x[0] < st.hT:
            self.conT1.set_color('blue')
            self.tube1.set_height(x[0] * self.hT / st.hT)
        elif x[0] >= st.hT:
            self.conT1.set_color('blue')
            self.tube1.set_height(self.hT)
        else:
            self.conT1.set_color('white')
            self.tube1.set_height(0)
        if 0 < x[1] < st.hT:
            self.conT1T2.set_color('blue')
            self.conT2.set_color('blue')
            self.tube2.set_height(x[1] * self.hT / st.hT)
        elif x[1] >= st.hT:
            self.conT1.set_color('blue')
            self.tube1.set_height(self.hT)
        else:
            self.conT1T2.set_color('white')
            self.conT2.set_color('white')
            self.tube2.set_height(0)
        self.canvas.draw()


pm.register_visualizer(MplTankVisualizer)



from collections import OrderedDict
import os
import pickle
import warnings

from scipy.integrate import ode
import sympy as sp
import numpy as np

from .simulation_modules import (
    Model, Solver, SolverException, Trajectory, TrajectoryException, Controller,
    Feedforward, SignalMixer, ModelMixer, ObserverMixer, Limiter, Sensor,
    Disturbance
)
from .controltools import calc_prefilter, place_siso

__all__ = ["LinearStateSpaceModel", "ODEInt", "ModelInputLimiter",
           "Setpoint", "HarmonicTrajectory", "SmoothTransition",
           "PIDController", "LinearStateSpaceController",
           "DeadTimeSensor", "GaussianNoise",
           "AdditiveMixer"]

"""
Ready to go implementations of simulation modules.
"""


class LinearStateSpaceModel(Model):
    """
    The state space model of a linear system.

    The parameters of this model can be provided in form of a file whose path is
    given by the setting ``config file`` .
    This path should point to a pickled dict holding the following keys:

        `system`:
            An Instance of :py:class:`scipy.signal.StateSpace` (from scipy)
            representing the system,
        `op_inputs`:
            An array-like object holding the operational point's inputs,
        `op_outputs`:
            An array-like object holding the operational point's outputs.

    """
    public_settings = OrderedDict([
        ("config file", None),
        ("initial state", None),
        ("initial output", None),
    ])

    def __init__(self, settings):
        file = settings["config file"]
        assert os.path.isfile(file)

        with open(file, "rb") as f:
            data = pickle.load(f)

        if "system" not in data:
            raise ValueError("Config file lacks mandatory settings.")

        self.ss = data["system"]

        # no feedthrough possible
        np.testing.assert_array_equal(self.ss.D, np.zeros_like(self.ss.D))

        settings["state_count"] = self.ss.B.shape[0]
        settings["input_count"] = self.ss.B.shape[1]

        self.input_offset = data.get("op_inputs",
                                     np.zeros((self.ss.B.shape[1], )))
        if len(self.input_offset) != self.ss.B.shape[1]:
            raise ValueError("Provided input offset does not match input "
                             "dimensions.")

        self.output_offset = data.get("op_outputs",
                                      np.zeros((self.ss.C.shape[0], )))
        if len(self.output_offset) != self.ss.C.shape[0]:
            raise ValueError("Length of provided output offset does not match "
                             "output dimensions ({} != {}).".format(
                len(self.output_offset),
                self.ss.C.shape[0]
            ))

        if settings["initial state"] is None:
            if settings["initial output"] is None:
                raise ValueError("Neither 'initial state' nor 'initial output'"
                                 "given.")

            settings["initial state"] = \
                np.linalg.pinv(self.ss.C) @ (settings["initial output"]
                                             - self.output_offset)

        super().__init__(settings)

    def state_function(self, t, x, args):
        return np.squeeze(self.ss.A @ x + self.ss.B @ (
            args[0] - self.input_offset))

    def calc_output(self, input_vector):
        return (self.ss.C @ input_vector
                + self.output_offset)


class ODEInt(Solver):
    """
    Wrapper for ode_int from Scipy project
    """
    public_settings = OrderedDict([
        ("Mode", "vode"),
        ("Method", "adams"),
        ("measure rate", 500),
        ("step size", 1e-3),
        ("rTol", 1e-6),
        ("aTol", 1e-9),
        ("start time", 0),
        ("end time", 5)
    ])

    def __init__(self, settings):
        Solver.__init__(self, settings)

        # setup solver
        if hasattr(self._model, "jacobian"):
            self._solver = ode(self._model.state_function,
                               jac=self._model.jacobian)
        else:
            self._solver = ode(self._model.state_function)

        self._solver.set_integrator(self._settings["Mode"],
                                    method=self._settings["Method"],
                                    rtol=self._settings["rTol"],
                                    atol=self._settings["aTol"],
                                    max_step=self._settings["step size"]
                                    )
        self._solver.set_initial_value(np.atleast_1d(self._model.initial_state),
                                       t=self._settings["start time"])

    @property
    def t(self):
        return self._solver.t

    @property
    def successful(self):
        return self._solver.successful()

    def set_input(self, *args):
        """
        propagate input changes to ode_int
        """
        self._solver.set_f_params(args)
        if hasattr(self._model, "jacobian"):
            # TODO Test
            self._solver.set_jac_params(args)

    def integrate(self, t):
        """
        Integrate until target step is reached.

        :param t: target time
        :return: system state at target time
        """
        state = self._solver.integrate(t + self._settings["step size"])

        # check model constraints
        new_state = self._model.root_function(state)
        if new_state[0]:
            # reset solver since discontinuous change in equations happened
            self._solver.set_initial_value(new_state[1], self.t)

        if not self._solver.successful():
            raise SolverException("Integration has not been successful.")

        return state


class SmoothTransition(Trajectory):
    """
    provides (differential) smooth transition between two scalar states
    """
    public_settings = {"states": [[0, 1]],
                       "start time": 0,
                       "delta t": 5,
                       }

    def __init__(self, settings):
        settings["states"] = np.asarray(settings["states"])
        Trajectory.__init__(self, settings)

        # setup symbolic expressions
        tau, k = sp.symbols("tau, k")

        gamma = self._settings["differential_order"] + 1
        alpha = sp.factorial(2 * gamma + 1)

        f = sp.binomial(gamma, k) * (-1) ** k * tau ** (gamma + k + 1) / (gamma + k + 1)
        phi = alpha / sp.factorial(gamma) ** 2 * sp.summation(f, (k, 0, gamma))

        # differentiate phi(tau), index in list corresponds to order
        dphi_sym = [phi]  # init with phi(tau)
        for order in range(self._settings["differential_order"]):
            dphi_sym.append(dphi_sym[-1].diff(tau))

        # lambdify
        self.dphi_num = []
        for der in dphi_sym:
            self.dphi_num.append(sp.lambdify(tau, der, "numpy"))

        # issue deprecation warning
        if self._settings["states"].ndim == 1:
            msg = "SmoothTransition will require 2d data for 'states' " \
                  "in the next version, please update your configuration."
            warnings.warn(msg, DeprecationWarning)
            self._logger.warn(msg)
            self._settings["states"] = self._settings["states"][None, :]

    def _desired_values(self, t):
        """
        Calculates desired trajectory
        """
        yd = self._settings['states']
        t0 = self._settings['start time']
        dt = self._settings['delta t']

        y = np.zeros((yd.shape[0], len(self.dphi_num)))
        if t < t0:
            y[:, 0] = yd[:, 0]
        elif t > t0 + dt:
            y[:, 0] = yd[:, 1]
        else:
            for order, dphi in enumerate(self.dphi_num):
                if order == 0:
                    ya = yd[:, 0]
                else:
                    ya = np.zeros_like(yd[:, 0])

                scale = dphi((t - t0) / dt) * 1 / dt ** order
                y[:, order] = ya + (yd[:, 1] - yd[:, 0]) * scale

        return y


class HarmonicTrajectory(Trajectory):
    """
    This generator provides a scalar harmonic sinus signal
    with derivatives up to order n
    """

    public_settings = OrderedDict([("Amplitude", 0.25),
                                   ("Frequency", 0.1),
                                   ("Offset", 0.75),
                                   ("Phase in degree", 0)])

    def __init__(self, settings):
        Trajectory.__init__(self, settings)

        # calculate symbolic derivatives up to order n
        t, a, f, off, p = sp.symbols("t, a, f, off, p")
        self.yd_sym = []
        harmonic = a * sp.sin(2 * sp.pi * f * t + p) + off

        for idx in range(settings["differential_order"] + 1):
            self.yd_sym.append(harmonic.diff(t, idx))

        # lambdify
        for idx, val in enumerate(self.yd_sym):
            self.yd_sym[idx] = sp.lambdify((t, a, f, off, p), val, "numpy")

    def _desired_values(self, t):
        # yd = []
        yd = np.zeros((self._settings['differential_order'] + 1), )

        a = self._settings['Amplitude']
        f = self._settings['Frequency']
        off = self._settings["Offset"]
        p = self._settings["Phase in degree"] * np.pi / 180

        for idx, val in enumerate(self.yd_sym):
            yd[idx] = val(t, a, f, off, p)
            # yd.append(val(t, a, f, off, p))

        return yd


class Setpoint(Trajectory):
    """
    Provides setpoints for every output component.

    If the output is not scalar, just add more entries to the list.
    By querying the differential order from the controller (if available) the
    required derivatives are given.

    Note:
        Keep in mind that while this class provides zeros for all derivatives
        of the desired value, they actually strive to infinity for :math:`t=0` .
    """

    public_settings = OrderedDict([("Setpoint", [0])])

    def __init__(self, settings):
        Trajectory.__init__(self, settings)
        self.yd = np.zeros((len(self._settings["Setpoint"]),
                            self._settings["differential_order"] + 1))

        for idx, val in enumerate(self._settings["Setpoint"]):
            self.yd[idx, 0] = val

    def _desired_values(self, t):
        return self.yd


class LinearStateSpaceController(Controller):
    """
    A controller that is based on a state space model of a linear system.

    This controller needs a linear statespace model, just as the
    :py:class:`LinearStateSpaceModel` . The file provided in ``config file``
    should therefore contain a dict holding the entries: ``model``,
    ``op_inputs`` and ``op_outputs`` .

    If poles is given (differing from `None` ) the state-feedback will be
    computed using :py:func:`pymoskito.place_siso` .
    Furthermore an appropriate prefilter is calculated, which establishes
    stationary attainment of the desired output values.

    Note:
        If a SIMO or MIMO system is given, the control_ package as well as the
        slycot_ package are needed the perform the pole placement.


    .. _control: https://github.com/python-control/python-control
    .. _slycot: https://github.com/python-control/Slycot
    """

    public_settings = OrderedDict([
        ("input source", "system_state"),
        ("config file", None),
        ("poles", None),
    ])

    def __init__(self, settings):
        file = settings["config file"]
        assert os.path.isfile(file)

        with open(file, "rb") as f:
            data = pickle.load(f)

        if "system" not in data:
            raise ValueError("Config file lacks mandatory settings.")

        self.ss = data["system"]

        self.input_offset = data.get("op_inputs", None)
        self.output_offset = data.get("op_outputs", None)

        if self.input_offset is None:
            self.input_offset = np.zeros((self.ss.B.shape[1], ))
        if len(self.input_offset) != self.ss.B.shape[1]:
            raise ValueError("Provided input offset does not match input "
                             "dimensions.")

        if self.output_offset is None:
            self.output_offset = np.zeros((self.ss.C.shape[0], ))
        if len(self.output_offset) != self.ss.C.shape[0]:
            raise ValueError("Length of provided output offset does not match "
                             "output dimensions ({} != {}).".format(
                len(self.output_offset),
                self.ss.C.shape[0]
            ))

        # add specific "private" settings
        settings.update(input_order=0)
        settings.update(output_dim=self.ss.C.shape[0])
        settings.update(input_type=settings["input source"])
        super().__init__(settings)

        if settings.get("poles", None) is None:
            # pretty useless but hey why not.
            self.feedback = np.zeros((self.ss.B.shape[1], self.ss.A.shape[0]))
        else:
            if self.ss.B.shape[1] == 1:
                # save the control/slycot dependency
                self.feedback = place_siso(self.ss.A,
                                           self.ss.B,
                                           self.settings["poles"])
            else:
                import control
                self.feedback = control.place(self.ss.A,
                                              self.ss.B,
                                              self.settings["poles"])

        self.prefilter = calc_prefilter(self.ss.A, self.ss.B, self.ss.C,
                                        self.feedback)

    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        return (-self.feedback @ input_values
                + self.prefilter @ (trajectory_values[:, 0] -
                                    self.output_offset)
                + self.input_offset)


class PIDController(Controller):
    """
    PID Controller
    """
    public_settings = OrderedDict([("Kp", 700),
                                   ("Ki", 500),
                                   ("Kd", 200),
                                   ("output_limits", [0, 255]),
                                   ("input_state", [2]),
                                   ("tick divider", 1)])
    last_time = 0

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(input_order=0)
        settings.update(output_dim=len(self.public_settings["input_state"]))
        settings.update(input_type="system_state")
        Controller.__init__(self, settings)

        # define variables for data saving in the right dimension
        # (columns vectors)
        self.e_old = np.zeros((len(self._settings["input_state"]), ))
        self.integral_old = np.zeros((len(self._settings["input_state"]), ))
        self.last_u = np.zeros((len(self._settings["input_state"]), ))
        self.output = np.zeros((len(self._settings["input_state"]), ))

    def _control(self, time, trajectory_values=None, feedforward_values=None, input_values=None, **kwargs):
        # input abbreviations
        x = np.zeros((len(self._settings["input_state"]), ))
        for idx, state in enumerate(self._settings["input_state"]):
            x[idx] = input_values[int(state)]

        yd = trajectory_values

        # step size
        dt = time - self.last_time
        # save current time
        self.last_time = time

        if dt != 0:
            for i in range(len(x)):
                # error
                e = yd[i] - x[i]
                integral = e * dt + self.integral_old[i]
                if integral > self._settings["output_limits"][1]:
                    integral = self._settings["output_limits"][1]
                elif integral < self._settings["output_limits"][0]:
                    integral = self._settings["output_limits"][0]
                differential = (e - self.e_old[i]) / dt

                self.output[i] = (self._settings["Kp"] * e
                                  + self._settings["Ki"] * integral
                                  + self._settings["Kd"] * differential)

                if self.output[i] > self._settings["output_limits"][1]:
                    self.output[i] = self._settings["output_limits"][1]
                elif self.output[i] < self._settings["output_limits"][0]:
                    self.output[i] = self._settings["output_limits"][0]

                # save data for new calculation
                self.e_old[i] = e
                self.integral_old[i] = integral
            u = self.output
        else:
            u = self.last_u
        self.last_u = u
        return u


class AdditiveMixer(SignalMixer):
    """
    Signal Mixer that accumulates all input signals.

    Processing is done according to rules of numpy broadcasting.
    """
    public_settings = OrderedDict([("Input A", None),
                                   ("Input B", None)])

    def __init__(self, settings):
        settings.update([("input signals", [settings["Input A"],
                                            settings["Input B"]])])
        SignalMixer.__init__(self, settings)

    def _mix(self, signal_values):
        return sum(signal_values)


class ModelInputLimiter(Limiter):
    """
    ModelInputLimiter that limits the model input values.

    Settings:
        `Limits`: (List of) list(s) that hold (min, max) pairs for the
        corresponding input.
    """

    public_settings = OrderedDict([("Limits", [None, None])])

    def __init__(self, settings):
        settings.update([("input_signal", "ModelMixer")])
        Limiter.__init__(self, settings)
        self.limits = np.atleast_2d(settings["Limits"])

    def _limit(self, values):
        val = np.atleast_1d(values)
        out = np.zeros_like(val)
        for idx, v in enumerate(val):
            lim = self.limits[idx]
            if lim[0] is None and lim[1] is None:
                out[idx] = v
            else:
                out[idx] = np.clip(v, *lim)

        if len(values.shape) == 1:
            return out.flatten()
        else:
            return out


class DeadTimeSensor(Sensor):
    """
    Sensor that adds a measurement delay on chosen states
    """

    public_settings = OrderedDict([("states to delay", [0]),
                                   ("delay", 1)])

    def __init__(self, settings):
        settings.update([("input signal", "system_state")])
        Sensor.__init__(self, settings)
        self._storage = None

    def _measure(self, value):
        if self._storage is None:
            # create storage with length "delay"
            # initial values are the first input
            self._storage = [value]*int(self._settings["delay"])

        # save current values
        measurement = value.copy()
        # add new measurement
        self._storage.append(value)

        # get delayed measurements
        delayed_measurement = self._storage.pop(0)

        # replace current values with delayed values, if it is chosen
        for i in self._settings["states to delay"]:
            measurement[i] = delayed_measurement[i]

        return measurement


class GaussianNoise(Disturbance):
    """
    Noise generator for gaussian noise
    """

    public_settings = OrderedDict([("sigma", 1),
                                   ("mean", 0)])

    def __init__(self, settings):
        settings.update([("input signal", "Sensor")])
        Disturbance.__init__(self, settings)

    def _disturb(self, value):
        return np.random.normal(self._settings['mean'],
                                self._settings['sigma'],
                                value.output_dim)



# -*- coding: utf-8 -*-

import sympy as sp
import numpy as np
import matplotlib.pyplot as plt

import pymoskito as pm
import settings as st

from matplotlib import rcParams
rcParams['text.usetex'] = True


def preview(expr, **kwargs):
    """
    support function to display nice formula
    :param expr:
    :param kwargs:
    :return:
    """
    latex_str = sp.latex(expr, **kwargs)
    latex_str = latex_str.replace("operator_name", "mathrm")
    plt.text(0.1, 0.1, latex_str, fontsize=20)
    plt.axis('off')
    plt.show()

mat2array = [{'ImmutableMatrix': np.array}, 'numpy']
params = sp.symbols('x1, x2, x3, x4,'
                    'x1_d, x2_d, x3_d, x4_d, z,'
                    'xd_d0, xd_d1, xd_d2, xd_d3, xd_d4,'
                    'x1_e, x2_e, x3_e, x4_e, z_e,'
                    'u, l1, l2, l3, l1_star, l2_star, l3_star,'
                    'T, d, k_s, k_L, k_V, A_B, A_Sp, m, g,'
                    'x1_d_dot, x2_d_dot, x3_d_dot, x4_d_dot, x1_dot, c0, c1, c2')
x1, x2, x3, x4, \
    x1_d, x2_d, x3_d, x4_d, z, \
    xd_d0, xd_d1, xd_d2, xd_d3, xd_d4, \
    x1_e, x2_e, x3_e, x4_e, z_e, \
    u, l1, l2, l3, l1_star, l2_star, l3_star, \
    T, d, k_s, k_L, k_V, A_B, A_Sp, m, g, \
    x1_d_dot, x2_d_dot, x3_d_dot, x4_d_dot, x1_dot, c0, c1, c2 = params

dict_names = {x1_d: r'x_{1}^{d}',
              x2_d: r'x_{2}^{d}',
              x3_d: r'x_{3}^{d}',
              x4_d: r'x_{4}^{d}',
              }

x = sp.Matrix([x1, x2, x3, x4])
x_e = sp.Matrix([x1_e, x2_e, x3_e, x4_e])
f_x = sp.Matrix([x2, -x1 / T ** 2 - 2 * d * x2 / T,
                 x4,
                 (k_L * ((k_V * x1 - A_B * x4) / A_Sp) ** 2 - m * g) / m])
g_x = sp.Matrix([0, k_s / T ** 2, 0, 0])
h_x = sp.Matrix([x3])

# preview(f_x, mode='equation', mat_str="array", symbol_names=dict_names)
# preview(g_x, mode='equation', mat_str="array", symbol_names=dict_names)

# linearization
simplification_list = [(x1, A_Sp*sp.sqrt(m*g/k_L)/k_V),
                       (x2, 0),
                       (x3, x3_e),
                       (x4, 0),
                       (u, A_Sp*sp.sqrt(m*g/k_L)/(k_V*k_s))]
sys = f_x + g_x*u
A_lin = sys.jacobian(x)
A_lin = A_lin.subs(simplification_list)
B_lin = sys.jacobian(sp.Matrix([u]))
B_lin = B_lin.subs(simplification_list)
C_lin = h_x.jacobian(x)
C_lin = C_lin.subs(simplification_list)
D_lin = h_x.jacobian(sp.Matrix([u]))
D_lin = D_lin.subs(simplification_list)
# preview(A_lin, mode='equation', mat_str="array", symbol_names=dict_names)
# preview(B_lin, mode='equation', mat_str="array", symbol_names=dict_names)
# preview(C_lin, mode='equation', mat_str="array", symbol_names=dict_names)
# preview(D_lin, mode='equation', mat_str="array", symbol_names=dict_names)

x_e_list = [x3_e]
parameter_list = [T, d, k_s, k_L, k_V, A_B, A_Sp, m, g]
A_func = sp.lambdify((x_e_list + parameter_list), A_lin, modules=mat2array)
B_func = sp.lambdify((x_e_list + parameter_list), B_lin, modules=mat2array)
C_func = sp.lambdify((x_e_list + parameter_list), C_lin, modules=mat2array)
D_func = sp.lambdify((x_e_list + parameter_list), D_lin, modules=mat2array)

A_num = A_func(0, st.T, st.d, st.k_s, st.k_L, st.k_V, st.A_B, st.A_Sp, st.m, st.g)
B_num = B_func(0, st.T, st.d, st.k_s, st.k_L, st.k_V, st.A_B, st.A_Sp, st.m, st.g)
C_num = C_func(0, st.T, st.d, st.k_s, st.k_L, st.k_V, st.A_B, st.A_Sp, st.m, st.g)
D_num = D_func(0, st.T, st.d, st.k_s, st.k_L, st.k_V, st.A_B, st.A_Sp, st.m, st.g)

# check controlability for state feedback
Q_c = np.concatenate((B_num,
                      A_num @ B_num,
                      np.linalg.matrix_power(A_num, 2) @ B_num,
                      np.linalg.matrix_power(A_num, 3) @ B_num), axis=1)
n = np.rank(Q_c)  # system is not controllable, because n=2 < 4 system order

# check controlability for state feedback with I-controller
A_cross = np.concatenate((np.concatenate((A_num, C_num), axis=0),
                          np.zeros((A_num.shape[0] + C_num.shape[0],
                                    C_num.shape[0]))),
                         axis=1)
B_cross = np.concatenate((B_num, np.zeros((C_num.shape[0], B_num.shape[1]))),
                         axis=0)
C_cross = np.concatenate((C_num, np.zeros((C_num.shape[0], C_num.shape[0]))),
                         axis=1)
Q_c_cross = np.concatenate((B_num,
                            A_num @ B_num,
                            np.linalg.matrix_power(A_num, 2) @ B_num,
                            np.linalg.matrix_power(A_num, 3) @ B_num,
                            np.linalg.matrix_power(A_num, 4) @ B_num),
                           axis=1)

# system is not controllable, because n=2 < 5 system order
n = np.rank(Q_c_cross)


# calculate lie derivatives
Lfh_n = pm.lie_derivatives(h_x, f_x, x, 4)
LgLfh = pm.lie_derivatives(Lfh_n[1], g_x, x, 1)
LgLf2h = pm.lie_derivatives(Lfh_n[2], g_x, x, 1)
LgLf3h = pm.lie_derivatives(Lfh_n[3], g_x, x, 1)

y_d0 = h_x
y_d1 = Lfh_n[1]
y_d2 = Lfh_n[2] + LgLfh
y_d3 = Lfh_n[3] + LgLf2h
y_d4 = Lfh_n[4] + LgLf3h

parameter = x1, x2, x3, x4, T, d, k_s, k_L, k_V, A_B, A_Sp, m, g
# y_d0_func = sp.lambdify(parameter, y_d0[0], modules=mat2array)  # trivial x3
# y_d1_func = sp.lambdify(parameter, y_d1[0], modules=mat2array)  # trivial x4
y_d2_func = sp.lambdify(parameter, y_d2[0], modules=mat2array)
y_d3_func = sp.lambdify(parameter, y_d3[0], modules=mat2array)

LgLf3h_func = sp.lambdify(parameter, LgLf3h[0], modules=mat2array)
Lf4h_func = sp.lambdify(parameter, Lfh_n[4][0], modules=mat2array)
f_func = sp.lambdify(parameter, f_x, modules=mat2array)
g_func = sp.lambdify(parameter, g_x, modules=mat2array)
h_func = sp.lambdify(parameter, h_x, modules=mat2array)

# some considerations about observability
# observability map
q = sp.Matrix([y_d0, y_d1, y_d2, y_d3])
# observability matrix
Q = q.jacobian(x)
Q_inv = sp.inv_quick(Q)
Q_inv_func = sp.lambdify(parameter, Q_inv, modules=mat2array)

# observer for the first subsystem (fan)
L_fan = sp.Matrix([[c1 - 2 * d / T],
                   [c0 - 2 * d * (c1 - 2 * d / T) / T - 1 / T ** 2]])
L_fan_func = sp.lambdify((c0, c1, T, d), L_fan, modules=mat2array)

f_fan_e = sp.Matrix([[x2_e],
                     [-x1_e / T ** 2 - 2 * d * x2_e / T + k_s * u / T ** 2]])
f_fan_e_func = sp.lambdify((x1_e, x2_e, T, d, k_s, u),
                           f_fan_e,
                           modules=mat2array)

# observer for the second subsystem (ball)
C = sp.Matrix([[1, 0, 0]])
C_dot = sp.Matrix([[0, 0, 0]])
C_ddot = sp.Matrix([[0, 0, 0]])
L = sp.Matrix([l1, l2, l3])
L_star = sp.Matrix([l1_star, l2_star, l3_star])

x = sp.Matrix([x3, x4, z])
x_e = sp.Matrix([x3_e, x4_e, z_e])
x_dot = sp.Matrix([x4,
                   (k_L * ((k_V * x1 - A_B * x4) / A_Sp) ** 2 - m * g + z) / m,
                   0])
x_dot_e = sp.Matrix([x4_e,
                     (k_L * ((k_V * x1 - A_B * x4_e) / A_Sp) ** 2
                      - m * g + z_e) / m,
                     0])
x_dot_obs_error = x_dot - x_dot_e

A = x_dot_obs_error.jacobian(x)
A = A.subs([(x4, x4_d), (x1, x1_d)])
A_dot = A.subs([(x4_d, x4_d_dot), (x1_d, x1_d_dot)])
Q_B = sp.Matrix([C,
                 C_dot + C * A,
                 C_ddot + C_dot * A + C * A_dot + C_dot * A + C * A ** 2])
Q_B_inv = sp.inv_quick(Q_B)

e_3 = sp.Matrix([0, 0, 1])
t_1 = Q_B_inv * e_3
T = sp.Matrix([t_1.transpose(),
               (A * t_1).transpose(),
               (A ** 2 * t_1).transpose()]).transpose()
T_inv = sp.inv_quick(T)
T_dot = sp.Matrix([[0, 0, 0],
                   [0, 0, T[1, 2]],
                   [0, 0, 0]])
T_dot = T_dot.subs([(x4_d, x4_d_dot),
                    (x1_d, x1_d_dot)])
A_star = T_inv * (A * T - T_dot)
A_star = A_star.expand()
C_star = C * T
lhs = A_star - L_star * C_star
A_d_star = sp.Matrix([[0, 0, -c0], [1, 0, -c1], [0, 1, -c2]])
equations = [sp.Eq(lhs[0, 2], A_d_star[0, 2]),
             sp.Eq(lhs[1, 2], A_d_star[1, 2]),
             sp.Eq(lhs[2, 2], A_d_star[2, 2])]
solution = sp.solve(equations, L_star)
L_star = sp.Matrix([solution[l1_star], solution[l2_star], solution[l3_star]])
L_ball = T * L_star
L_ball_func = sp.lambdify((c0, c1, c2, x1_d, x1_d_dot, x4_d, x4_d_dot, k_L, k_V,
                           A_B, A_Sp, m, g),
                          L_ball,
                          modules=mat2array)

f_ball_e = sp.Matrix([[x4_e],
                      [(k_L * ((k_V * x1 - A_B * x4_e) / A_Sp) ** 2
                        - m * g + z_e) / m],
                      [0]])
f_ball_e_func = sp.lambdify((x4_e, x1, z_e, k_L, k_V, A_B, A_Sp, m, g),
                            f_ball_e,
                            modules=mat2array)

x1_flat = (sp.sqrt((xd_d2 + g) * m * A_Sp ** 2 / k_L) + A_B * xd_d1) / k_V
x2_flat = (xd_d3 * m * A_Sp ** 2 / (2 * k_V * k_L * (k_V * x1_flat
                                                     - A_B * xd_d1))
           + A_B * xd_d2 / k_V)

x1_flat_func = sp.lambdify((xd_d1, xd_d2, k_L, k_V, A_B, A_Sp, m, g),
                           x1_flat,
                           modules=mat2array)
x2_flat_func = sp.lambdify((xd_d1, xd_d2, xd_d3, k_L, k_V, A_B, A_Sp, m, g),
                           x2_flat,
                           modules=mat2array)



# -*- coding: utf-8 -*-
import numpy as np
# -----------------------------------------------------------------------------
# global default settings for physical simulation
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# parameters of the tank system according to the

K = 2e-5                                # m**3 / (s V) - gain of pump

rT = 0.04                               # m - radius of the tanks
g = 9.81                                # m / s^2 - gravity

AT = np.pi * rT ** 2                    # m**2 - Tanks cross section
hT = 0.8                                # m - height of the tanks

rS1 = 0.0035                            # m - radius of pipe cross section between tank 1 and 2
rS2 = 0.0035                            # m - radius of pipe cross section after tank 2

AS1 = np.pi * rS1 ** 2                  # m**2 - Outlet cross section of drain between tank 1 and 2
AS2 = np.pi * rS2 ** 2                  # m**2 - Outlet cross section of drain of tank 2

poles_state = np.array([-2, -2])        # poles for the state controller
poles_obs = np.array([-0.05, -0.05])    # poles for the High-Gain observer

limits_ctrl = np.array([0, 5])          # limits for the controller
input_ctrl = np.array([1])              # input state for the controller

initial_states = np.array([0, 0])       # initial states for the model

Kp = 1125                               # Gain value for the proportional part of PID controller
Ti = 0.005                              # Time value for the integral part of the PID controller
Td = 0                                  # Time value for the derivation part of the PID controller
# -----------------------------------------------------------------------------
# export settings latex-style
# -----------------------------------------------------------------------------
latex_font_size = 14
label_size = 1 * latex_font_size
title_size = 1.5 * latex_font_size



# -*- coding: utf-8 -*-
""" simulator interface
    provides functions to manipulate settings of the simulator and
    to inspect its current state.
"""
import ast
import copy
import logging
import sys
from collections import OrderedDict

import numpy as np
from PyQt5.QtCore import (
    Qt, QObject, pyqtSignal, pyqtSlot, QModelIndex, QSize, QThread, )
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtWidgets import QItemDelegate, QComboBox, QTreeView

from . import simulation_modules
from .registry import (
    get_registered_simulation_modules, get_simulation_module_class_by_name
)
from .simulation_core import (
    Simulator, SimulationSettings, SimulationStateChange
)


class SimulatorModel(QStandardItemModel):
    def __init__(self, parent=None):
        QStandardItemModel.__init__(self, parent=parent)

    def flags(self, index):
        if index.column() == 1:
            return Qt.ItemIsEditable | Qt.ItemIsEnabled
        else:
            return Qt.ItemIsEnabled


class PropertyItem(QStandardItem):
    RawDataRole = Qt.UserRole + 1

    def __init__(self, data):
        QStandardItem.__init__(self)
        self._logger = logging.getLogger(self.__class__.__name__)
        self._data = data
        self._text = self._get_text(data)

    def type(self):
        return QStandardItem.UserType

    def _get_text(self, data):
        return str(data)

    def setData(self, Any, role=None, *args, **kwargs):
        if role == Qt.EditRole:
            try:
                self._data = ast.literal_eval(Any)
            except (SyntaxError, ValueError) as e:
                # print(e)
                self._logger.exception(e)
                return
            self._text = str(self._data)

        elif role == self.RawDataRole:
            self._data = Any
            self._text = self._get_text(Any)

        else:
            raise NotImplementedError

        self.emitDataChanged()

    def data(self, role=None, *args, **kwargs):
        if role == Qt.DisplayRole:
            return self._text
        elif role == Qt.EditRole:
            if isinstance(self._data, str):
                return "'" + self._text + "'"
            else:
                return self._text
        elif role == self.RawDataRole:
            return self._data

        else:
            return super().data(role, *args, **kwargs)


class PropertyDelegate(QItemDelegate):
    """
    A delegate that manages all property settings.
    For now it uses a combobox for simulationModules and a standard
    delegate for the rest.
    """

    def __init__(self, parent=None):
        QItemDelegate.__init__(self, parent)
        self.comboDel = ComboDelegate()

    def createEditor(self, parent, option, index):
        if index.parent().isValid():
            # item has a parent -> no SimulationModule
            # TODO implement dropdown menu if property is a dict
            return QItemDelegate.createEditor(self, parent, option, index)
        else:
            # no parent -> top of hierarchy
            return self.comboDel.createEditor(parent, option, index)

    def setEditorData(self, editor, index):
        if isinstance(editor, QComboBox):
            self.comboDel.setEditorData(editor, index)
        else:
            QItemDelegate.setEditorData(self, editor, index)

    def setModelData(self, editor, model, index):
        if isinstance(editor, QComboBox):
            self.comboDel.setModelData(editor, model, index)
        else:
            QItemDelegate.setModelData(self, editor, model, index)


class ComboDelegate(QItemDelegate):
    """
    A delegate that adds a combobox to cells that lists
    all available types of Subclasses of SimulationModule
    """

    def __init__(self, parent=None):
        QItemDelegate.__init__(self, parent)

    def createEditor(self, parent, option, index):
        editor = QComboBox(parent)
        editor.addItems(self.extract_entries(index))
        editor.currentIndexChanged.connect(self.current_index_changed)
        return editor

    def setEditorData(self, editor, index):
        name = index.model().itemFromIndex(index).text()
        editor.blockSignals(True)
        editor.setCurrentIndex(editor.findText(name))
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        model.setData(
            index,
            editor.currentText() if editor.currentText() != "None" else None,
            role=PropertyItem.RawDataRole)

    @pyqtSlot(int)
    def current_index_changed(self, idx):
        self.commitData.emit(self.sender())

    @staticmethod
    def extract_entries(index):
        """
        extract all possible choices for the selected SimulationModule
        """
        entries = ["None"]
        idx = index.model().index(index.row(), 0, QModelIndex())
        sim_module_name = str(index.model().itemFromIndex(idx).text())
        sim_module = getattr(simulation_modules, sim_module_name)
        sub_modules = get_registered_simulation_modules(sim_module)
        for sub_module in sub_modules:
            entries.append(sub_module[1])

        return entries


class SimulatorView(QTreeView):
    def __init__(self, parent=None):
        QTreeView.__init__(self, parent)
        self.setItemDelegateForColumn(1, PropertyDelegate(self))

    def sizeHint(self):
        return QSize(300, 150)

    def minimumSizeHint(self):
        return self.sizeHint()


class SimulatorInteractor(QObject):
    """
    Class that interacts between the gui which controls the programs execution
    and the Simulator which handles the time step simulation
    """

    # signals
    simulation_finalized = pyqtSignal(str, dict)
    simulationProgressChanged = pyqtSignal(int)

    def __init__(self, parent=None):
        QObject.__init__(self, parent)
        self._logger = logging.getLogger(self.__class__.__name__)

        self.last_progress = 0
        self.end_time = 0
        self._setup_model()

        self._worker = None
        self._sim_settings = None
        self.simThread = QThread()
        self._sim_modules = {}
        self._sim_data = None
        self._sim_state = None

    def _setup_model(self):
        # create model
        self.target_model = SimulatorModel(self)
        self.target_model.itemChanged.connect(self.item_changed)

        # insert header
        self.target_model.setHorizontalHeaderLabels(['Property', 'Value'])

        # insert items
        self._setup_model_items()

    def _setup_model_items(self):
        """
        fill model with items corresponding to all predefined SimulationModules
        """

        # build initialisation list
        setup_list = Simulator.module_list

        # push init of Trajectory module to the end because it needs the
        # derivative orders of controller and feedforward
        setup_list.append("Trajectory")
        setup_list.remove("Trajectory")

        # insert main items
        for sim_module in setup_list:
            name = PropertyItem(sim_module)
            value = PropertyItem(None)
            new_items = [name, value]
            self.target_model.appendRow(new_items)

        # insert settings
        for row in range(self.target_model.rowCount()):
            index = self.target_model.index(row, 0)
            self._add_settings(index)

    def _add_settings(self, index):
        parent = index.model().itemFromIndex(index)
        child = index.model().item(index.row(), 1)
        module_name = parent.data(role=PropertyItem.RawDataRole)
        sub_module_name = child.data(role=PropertyItem.RawDataRole)
        if sub_module_name is None:
            return

        settings = self._read_settings(module_name, sub_module_name)
        for key, val in settings.items():
            setting_name = PropertyItem(key)
            setting_value = PropertyItem(val)
            parent.appendRow([setting_name, setting_value])

    def _read_settings(self, module_name, sub_module_name):
        """
        Read the public settings from a simulation module
        """
        module_cls = getattr(simulation_modules, module_name)
        sub_module_cls = get_simulation_module_class_by_name(module_cls,
                                                             sub_module_name)
        return sub_module_cls.public_settings

    def item_changed(self, item):
        if item.parent():
            return

        idx = item.index()
        module_item = idx.model().item(idx.row())

        # delete all old settings
        module_item.removeRows(0, module_item.rowCount())

        # insert new settings
        self._add_settings(module_item.index())

        return

    def _get_settings(self, model, module_name):
        item = model.findItems(module_name).pop(0)

        # TODO this is not the good way --> switch to pyqtgraph's implementation
        settings = OrderedDict()
        for row in range(item.rowCount()):
            property_name = self.target_model.data(item.child(row, 0).index(),
                                                   role=PropertyItem.RawDataRole)
            property_val = self.target_model.data(item.child(row, 1).index(),
                                                  role=PropertyItem.RawDataRole)
            settings.update({property_name: property_val})

        return settings

    def _setup_sim_modules(self, model):
        """
        Setup simulation Modules.

        :param model: model holding the public settings of each module

        Returns:
            bool: If setup was successful.
        """
        for row in range(model.rowCount()):
            # build correct object and add it to the simulator
            module_item = model.item(row, 0)
            module_name = str(module_item.text())
            sub_module_item = model.item(row, 1)
            sub_module_name = str(sub_module_item.text())

            if sub_module_name == 'None':
                continue

            # get class
            module_cls = getattr(simulation_modules, module_name)
            sub_module_cls = get_simulation_module_class_by_name(
                module_cls,
                sub_module_name)

            # get public settings for module
            settings = self._get_settings(self.target_model, module_item.text())
            if settings is None:
                return False
            settings.update({"type": sub_module_name})
            settings.update({"modules": self._sim_modules})

            # append special settings
            if module_name == "Solver":
                if "Model" not in settings["modules"]:
                    return False
                self._sim_settings = SimulationSettings(
                    settings["start time"],
                    settings["end time"],
                    settings["step size"],
                    settings["measure rate"])

            # build object
            try:
                slot = sub_module_cls(settings)
            except Exception as e:
                self._logger.error("Init of module '{}' failed with\n"
                                   "'{}'!".format(module_name, e))
                return False

            # add to simulation modules
            self._sim_modules.update({module_name: slot})

            # store settings
            # self._sim_data['modules'].update({module_name: settings})

        if all([mod in self._sim_modules
                for mod in Simulator.static_module_list]):
            return True

        return False

    def set_regime(self, reg):
        """
        Load the given regimes settings into the target model.

        Returns:
            bool: `True` if successful, `False` if errors occurred.
        """
        if reg is None:
            return
        if isinstance(reg, list):
            self._logger.error("setRegime(): only scalar input allowed!")
            return False

        return self._apply_regime(reg, False)

    def restore_regime(self, reg):
        """
        Restore the given generated regime settings into the target model.

        Returns:
            bool: `True` if successful, `False` if errors occurred.
        """
        if reg is None:
            return
        if isinstance(reg, list):
            self._logger.error("restoreRegime(): only scalar input allowed!")
            return False

        return self._apply_regime(reg, True)

    def _apply_regime(self, reg, ignore_is_public):
        """
        Set all module settings to those provided in the regime.

        Returns:
            bool: `True` if successful, `False` if errors occurred.
        """
        if reg["clear previous"]:
            # delete all items
            self.target_model.removeRows(0, self.target_model.rowCount())

            # load module defaults
            self._setup_model_items()

        # overwrite all settings with the provided ones
        for module_name, value in reg.items():
            if module_name == "Name" or module_name == "clear previous":
                continue

            # sanity check
            module_cls = getattr(simulation_modules, module_name, None)
            if module_cls is None:
                self._logger.error("_apply_regime(): No module called {0}"
                                   "".format(module_name))
                return False

            items = self.target_model.findItems(module_name)
            if not len(items):
                self._logger.error("_apply_regime(): No item in List called {0}"
                                   "".format(module_name))
                return False

            module_item = items.pop(0)
            module_type = value["type"]

            # sanity check
            sub_module_cls = get_simulation_module_class_by_name(module_cls,
                                                                 module_type)

            if not sub_module_cls:
                self._logger.error("_apply_regime(): No sub-module called {0}"
                                   "".format(module_type))
                return False

            module_index = module_item.index()
            module_type_index = module_index.model().index(module_index.row(),
                                                           1)
            module_index.model().setData(module_type_index,
                                         module_type,
                                         role=PropertyItem.RawDataRole)
            # due to signal connections, default settings are loaded
            # automatically in the back

            # overwrite specific settings
            for key, val in value.items():
                if key == "type":
                    continue

                for row in range(module_item.rowCount()):
                    if self.target_model.data(
                            module_item.child(row, 0).index()) == key:
                        value_idx = self.target_model.index(row,
                                                            1,
                                                            module_index)
                        self.target_model.setData(value_idx,
                                                  val,
                                                  role=PropertyItem.RawDataRole)
                        break
                else:
                    if not ignore_is_public:
                        self._logger.error(
                            "_applyRegime(): No public setting called '{0}'"
                            "available for Module: '{1}'".format(key,
                                                                 module_type))
                        return False

        return True

    @pyqtSlot()
    def run_simulation(self):
        """
        entry hook for time step simulation
        - use settings to create modules in simulation loop
        - move them into an extra thread
        - start simulation
        """
        # setup simulation modules
        suc = self._setup_sim_modules(self.target_model)
        if not suc:
            self._logger.error("Simulation Setup failed. Check Configuration.")
            self.simulation_finalized.emit("aborted", dict())
            return

        # setup simulator
        self._worker = Simulator(self._sim_settings, self._sim_modules)
        self._worker.moveToThread(self.simThread)

        # setup simulation modules
        for _module in self._sim_modules.values():
            _module.moveToThread(self.simThread)

        # setup signal connections
        self.simThread.started.connect(self._worker.run)
        self._worker.state_changed.connect(self.simulation_state_changed)
        self._worker.work_done.connect(self.simThread.quit)
        self.simThread.finished.connect(self.thread_finished)
        self.end_time = self._sim_settings.end_time

        # run
        self.simThread.start()

    @pyqtSlot(SimulationStateChange)
    def simulation_state_changed(self, state_change):
        """
        slot for simulation state changes

        :param state_change: see :cls:SimulationStateChange
        """
        self._logger.debug("simulation state change '{}'".format(
            state_change.type))

        if state_change.type == "start":
            self._sim_state = "running"

        elif state_change.type == "time":
            self._logger.debug("Reached simulation time {0}".format(
                state_change.t))
            progress = int(state_change.t / self._sim_settings.end_time * 100)
            if progress != self.last_progress:
                self._logger.info("Simulation reached {0}%".format(progress))
                self.simulationProgressChanged.emit(progress)
                self.last_progress = progress

        elif state_change.type == "abort":
            self._sim_state = "aborted"
            self._sim_data = copy.deepcopy(state_change.data)
            if isinstance(state_change.info, str):
                self._logger.info(state_change.info)
            else:
                self._logger.error("Simulation has been aborted due to an "
                                   "exception",
                                   exc_info=state_change.info)
                self._logger.warning("check your configuration")

        elif state_change.type == "finish":
            self._sim_state = "finished"
            self._sim_data = copy.deepcopy(state_change.data)

        else:
            self._logger.error(
                "simulation_state_changed(): ERROR Unknown state {0}".format(
                    state_change.type))

    @pyqtSlot()
    def stop_simulation(self):
        self._worker.stop()

    def _sim_aftercare(self):
        # reset internal states
        self._sim_settings = None

        # delete modules
        for module in self._sim_modules.keys():
            del module
        self._sim_modules = {}

        # don't disconnect signals in debug-mode
        if sys.gettrace() is None:
            self.simThread.started.disconnect(self._worker.run)
            self._worker.state_changed.disconnect(self.simulation_state_changed)
            self._worker.work_done.disconnect(self.simThread.quit)
            self.simThread.finished.disconnect(self.thread_finished)

        # delete simulator
        self._logger.info("deleting simulator")
        del self._worker

    def _postprocessing(self):
        """
        calculation of some basic metrics for quick judging of simulation results
        """

        # TODO make the entries to compare selectable
        # control and observer error
        if "Trajectory" in self._sim_data["results"]:
            m_data = self._get_result_by_name("Model")
            t_data = self._get_result_by_name("Trajectory")
            if len(t_data.shape) == 2:
                c_error = m_data[..., 0] - t_data[:, 0]
            elif len(t_data.shape) == 3:
                c_error = np.array([m_data[:, idx] - t_data[:, idx, 0]
                                    for idx in range(m_data.shape[1])]).T
            else:
                raise ValueError("Unknown Trajectory Format.")
            self._sim_data["results"].update(control_error=c_error)

        if "Observer" in self._sim_data["results"]:
            o_error = (self._get_result_by_name("Observer")
                       - self._get_result_by_name("Solver"))
            self._sim_data["results"].update(observer_error=o_error)

    def _get_result_by_name(self, name):
        return self._sim_data["results"][name]

    @pyqtSlot()
    def thread_finished(self):
        """
        Slot to be called when the simulation thread is finished.
        """
        assert self._sim_state == "finished" or self._sim_state == "aborted"

        # TODO move this somewhere else and support vectorial block outputs
        # Add some error prone calculations
        # self._postprocessing()

        # Rest internal states and disconnect signal connections
        self._sim_aftercare()

        # Signal gui that new data is available
        self.simulation_finalized.emit(self._sim_state, self._sim_data)
